#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "d2stats.h"

int logging = 3;

/* Writes debug information to stderr, can be adjusted by changing global
   'logging' above */
void write_log(const char *text, const unsigned int log_lvl, ...) {
  va_list args;
  if((unsigned int)logging >= log_lvl) {
    va_start(args, log_lvl);
    vfprintf(stderr, text, args);
    va_end(args);
  }
}

/* Writes 'len' zeros to the block of memory at 'bytes'. */
void zero_out(void *bytes, const unsigned int len) {
  unsigned int pos;
  for(pos = 0; pos < len; pos++) {
    ((uint8 *)bytes)[pos] = 0;
  }
}

/* Returns a mirror image of 'in', reversing 'len' (at most 32) bits.
   e.g., bitreverse(001101, 6) == 101100 */
uint32 bitreverse(const uint32 in, const unsigned int len) {
  unsigned int pos;
  uint32 out = 0;
  for(pos = 0; pos < len; pos++) {
    out += (in >> pos & 1) << (len - pos - 1);
  }
  return out;
}

/* Returns the first 'bitlen' bits (at most 32) at 'bytes', using 'bitpos'
   as a number-of-bits offset. Can break byte boundary if necessary.
   e.g., extract_int(bytes, 32, 8) will extract a char from bytes[4]. */
uint32 extract_int(const uint8 *bytes, const unsigned int bitpos, const unsigned int bitlen) {
  unsigned int pos;
  uint32 out = 0;
  for(pos = bitpos; pos < bitlen + bitpos; pos++) {
    out += (bytes[pos / 8] >> (pos % 8) & 1) << (pos - bitpos);
  }
  return out;
}

/* Returns the first 'bitlen' bits (at most 32) at 'bytes', using 'bitpos'
   as a number-of-bits offset, just like extract_int(), but also bitreverses
   the return val (just like bitreverse() would do). Can break byte boundary.
   e.g., bytes[1] = 001101; extract_rev_int(bytes, 8, 6) == 101100 */
uint32 extract_rev_int(const uint8 *bytes, const unsigned int bitpos, const unsigned int bitlen) {
  unsigned int pos;
  uint32 out = 0;
  for(pos = bitpos; pos < bitlen + bitpos; pos++) {
    out += (bytes[pos / 8] >> (7 - pos % 8) & 1) << (pos - bitpos);
  }
  return out;
}

/* Extracts 'len' bytes from 'pos' in 'charfile', placing them in 'bytes'.
   'bytes' must point to a memory block that's as least as long as 'len',
   'charfile' must be open and its seek position can change.
   Returns number of bytes successfully extracted */
unsigned int read_bytes(FILE *charfile, uint8 *bytes, const unsigned int pos, const unsigned int len) {
  if(fseek(charfile, pos, SEEK_SET)) {
    return 0; }
  return (unsigned int)fread(bytes, sizeof(uint8), len, charfile);
}

/* Returns the size of a file handle, 'file'. Will change its seek pos. */
unsigned int get_filesize(FILE *file) {
  unsigned int out;
  if(fseek(file, 0, SEEK_END)) {
    return 0; }
  return (out = ftell(file)) == -1 ? 0 : out;
}

/* Returns the number of times cstring 'small' occurs in cstring 'big'. */
unsigned int find_str(char *big, char *small) {
  unsigned int out = 0, pos;
  for(pos = 0; pos < strlen(big) - strlen(small); pos++) {
    if(!strncmp(big + pos, small, strlen(small))) {
      out++; }
  }
  return out;
}

/* Returns the number of set bits for 'in'. */
unsigned int count_bits(unsigned int in) {
  unsigned int pos, bits = 0;
  for(pos = 0; pos < sizeof(unsigned int) * 8; pos++) {
    bits += in >> pos & 1;
  }
  return bits;
}

/* Removes whitespace from the start and end of a cstring. Returns a new
   pointer to cstring 'in'. */
char *trim(char *in) {
  int pos, start, end;
  if(in[0] == '\0') {
    return in; }
  for(pos = 0; pos < (int)strlen(in) && in[pos] == ' '; pos++);
  start = pos;
  for(pos = (int)strlen(in) - 1; pos >= 0 && in[pos] == ' '; pos--);
  end = pos;
  in += start;
  in[end + 1] = '\0';
  return in;
}

/* Removes whitespace, copies the resulting cstring, frees the old one.
   Returns new pointer to new cstring. */
char *trim_realloc(char *in) {
  char *out, *tmp = trim(in);
  if(!(out = malloc(strlen(tmp) + 1))) {
    return NULL; }
  strncpy(out, tmp, strlen(tmp) + 1);
  free(in);
  return out;
}

/* Prints a length of time, in seconds, to buf. */
void print_time(const unsigned int in, char *buf) {
  const unsigned int days = in / (3600 * 24);
  const unsigned int hours = (in - days * 3600 * 24) / 3600;
  const unsigned int minutes = (in - hours * 3600) / 60;
  const unsigned int seconds = in % 60;
  /* Although this function could just spit out '1d5h30m2s' or something
     I decided to make this a little more compact and easy-to-read. */
  if(days) {
    sprintf(buf, "%d days, %d hours", days, hours); }
  else if(hours) {
    sprintf(buf, "%d:%02d", hours, minutes); }
  else if(minutes) {
    sprintf(buf, "%d minutes", minutes); }
  else {
    sprintf(buf, "%d seconds", seconds); }
}

/* Using a seconds-since-epoch time value, prints out the time of day
   in a format like '4:28 pm'. Not used anymore. */
void print_timeofday(const unsigned int in, char *buf) {
  struct tm *ptr;
  time_t tmp = in;
  unsigned int hours, minutes, hour;
  enum flag_t is_pm;
  ptr = localtime(&tmp);
  hours = ptr->tm_hour;
  minutes = ptr->tm_min;
  if(hours >= 12) {
    is_pm = yes; }
  if(hours == 0) {
    hour = 12; }
  else if(hours > 13) {
    hour = hours - 12; }
  else {
    hour = hours; }
  sprintf(buf, "%02d:%02d %s", hour, minutes, is_pm == yes ? "pm" : "am");
}
