#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifdef WIN32
#include "dirent.h"
#else
#include <dirent.h>
#endif

#include "d2stats.h"
#include "xml.h"
#include "tally.h"

/* Reads a Diablo II 1.10/1.11/1.12 expansion character save file and mallocs
   a charinfo_t entry in 'characters' at 'char_no' to store the data.
   Will not extract item data (time-consuming) if the file does not pass
   the user-specified character filter. (ladder, hardcore, etc.)
   Returns 0 if nothing major went wrong, 1 if there's a file or memory error,
   2 if the charsave file doesn't appear to be vaild. */
unsigned int read_file(char *filename, struct charinfo_t **characters, const unsigned int char_no) {
  FILE *charfile;
  struct charinfo_t *charinfo;
  if(!(characters[char_no] = malloc(sizeof(struct charinfo_t)))) {
    printf("d2stats: malloc failed (out of memory?)\n");
    return 1; }
  charinfo = characters[char_no];
  zero_out(charinfo, sizeof(struct charinfo_t));
  charinfo->filename = filename;
  charinfo->acct = ""; /* for now, this isn't ever set to anything else */
  if(!(charfile = fopen(filename, "rb"))) {
    write_log("could not open file %s\n", 3, filename);
    return 1; }
  if(!(charinfo->filesize = get_filesize(charfile))) {
    write_log("could not get filesize\n", 3);
    return 1; }
  if(charinfo->filesize == 130) { /* size of 'newbie.save' */
    write_log("%s is a new character file, there's nothing to display.\n", 4, filename);
    fclose(charfile);
    return 0; }
  /* most Diablo II files are 2-4kb at most */
  if(charinfo->filesize > 16 * 1024 || charinfo->filesize < 512) {
    write_log("file %s size abnormal\n", 3, filename);
    fclose(charfile);
    return 2; }
  if(!(charinfo->stats_len = extract_stats(charfile, &charinfo->stats))) {
    write_log("could not extract stats for file %s\n", 3, filename);
    fclose(charfile);
    return 2; }
  if(!(extract_class(charfile, charinfo))) {
    write_log("could not extract class for file %s\n", 3, filename);
    fclose(charfile);
    return 2; }
  if(!(extract_name(charfile, charinfo))) {
    write_log("could not extract character name for file %s\n", 3, filename);
    fclose(charfile);
    return 2; }
  if(!(extract_weapon_set(charfile, charinfo))) {
    write_log("d2stats: could not extract weapon set byte for file %s\n", 3, filename);
    return 2; }
  if(!(extract_title(charfile, charinfo))) {
    write_log("d2stats: could not extract title byte for file %s\n", 3, filename);
    return 2; }
  if(!(extract_chartype(charfile, charinfo))) {
    write_log("d2stats: character stored in file %s is likely non-expansion\n", 4, filename);
    return 2; }
  if(!(extract_res(charfile, charinfo))) {
    write_log("d2stats: bad quest info for file %s\n", 3, filename);
    return 2; }
  if(passes_filter(charinfo) == no) {
    fclose(charfile);
    return 0; }
  if(!(extract_item_sections(charfile, charinfo))) {
    write_log("d2stats: could not extract items for \"%s\"\n", 3, charinfo->name);
    fclose(charfile);
    return 2; }
  fclose(charfile);
  return 0;
}

/* Uses config var 'mode' to determine how to process 'charinfo'. Might either
   output the character data to XML, or tally equipped bonuses for ranking
   purposes. */
void process_file(struct charinfo_t *charinfo, struct list_t *list) {
  static char buf[1024] = "";
  if(buf[0] == '\0') {
    config(config_mode, buf); }
  if(!strncmp(buf, "character", 10)) {
    xml_print(charinfo); }
  else if(!strncmp(buf, "stats", 6)) {
    tally_items(charinfo, list); }
}

/* Using an array of cstrs, 'files', reads 'file_count' character save files
   into an equally large, previously malloc'd block of charinfo_t structs,
   'charaters'. Frees item data and total proplists, but first processes each
   character file. Returns what read_file() last returned (besides zero,
   unless no errors) and also assigns 'char_count' with the actual number of 
   valid character files that passed filtering. */
unsigned int read_files(char **files, const unsigned int file_count, struct charinfo_t **characters, struct list_t *list, unsigned int *char_count) {
  unsigned int file_no, curr_val;
  int return_val = 0;
  *char_count = 0;
  for(file_no = 0; file_no < file_count; file_no++) {
    curr_val = read_file(files[file_no], characters, file_no);
    if(!curr_val && characters[file_no]->items) {
      process_file(characters[file_no], list); 
      free_items(characters[file_no]);
      free_total(characters[file_no]->total);
      (*char_count)++;
    }
    else {
      if(curr_val != return_val) {
	return_val = curr_val; }
    }
  }
  return return_val;
}

/* Prepares the list for printing, and then prints it in XML format. */
void list_files(struct charinfo_t **characters, struct list_t *list, const unsigned int file_count) {
  parse_accounts(list);
  parse_games(list);
  link_ears_to_owners(list, characters, file_count);
  rank(list, characters, file_count);
  xml_list_print(list);
}

/* Frees character data and the accompanying cstring array with filenames. */
void free_files(struct charinfo_t **characters, char **files, const unsigned int file_count) {
  unsigned int file_no;
  for(file_no = 0; file_no < file_count; file_no++) {
    free(characters[file_no]);
    free(files[file_no]);
  }
  free(characters);
  free(files);
}

/* Uses argv / argc and various config variables to determine which charsave
   files the user wants to process. Returns the amount of filenames and
   assigns 'out_files' with the actual cstring array, which must be freed. */
unsigned int expand_args(int argc, char **argv, char ***out_files) {
  unsigned int count, file_no, limit, extra, dirname_len;
  DIR *dir;
  char **files, *name, *dirname, buf[1024] = "";
  struct dirent *dp;
  enum flag_t has_extra_slash = no;
  count = get_files(argc, argv, &files);
  /* if there are no files specified, add config var 'charsave' to the file
     list, but first check to make sure it's an openable directory. */
  config(charsave_path, buf);
  if(count == 0 && buf[0] != '\0') {
    if((dir = opendir(buf))) {
      closedir(dir);
      if(!(files = malloc(sizeof(char *)))) {
	return 0; }
      files[0] = buf;
      count++; }
    else {
      write_log("could not open default charsave path \"%s\"\n", 2, buf); }
  }
  for(file_no = 0; file_no < count; file_no++) {
    if((dir = opendir(files[file_no]))) { /* check if file is a directory. */
      dirname = files[file_no];
      dirname_len = (unsigned int)strlen(dirname);
      if(dirname_len > 1 && dirname[dirname_len - 1] == '/') {
	has_extra_slash = yes; }
      /* check how many files are in the dir */
      for(extra = 0, dp = readdir(dir); dp; dp = readdir(dir)) {
	if(strncmp(dp->d_name, ".", 2) && strncmp(dp->d_name, "..", 3)) {
	  extra++; }
      }
      closedir(dir);
      /* resize the array so that there's enough room for the files in this
         directory. overwrite the directory name itself */
      if(!(files = realloc(files, (count + extra - 1) * sizeof(char *)))) {
	return 0; }
      count += extra - 1;
      dir = opendir(dirname);
      for(limit = file_no + extra, dp = readdir(dir); dp && file_no < limit; dp = readdir(dir)) {
	if(strncmp(dp->d_name, ".", 2) && strncmp(dp->d_name, "..", 3)) {
	  /* need enough room for "path/to/dir" + extra slash if necessary
	     + filename + 1 for null term. */
	  if(!(name = malloc(dirname_len + strlen(dp->d_name) + (has_extra_slash == yes ? 1 : 2)))) {
	    return 0; }
	  strncpy(name, dirname, dirname_len + 1);
	  if(has_extra_slash == no) {
	    strncat(name, DIR_SLASH, 2); }
	  strncat(name, dp->d_name, strlen(dp->d_name) + 1);
	  files[file_no] = name;
	  file_no++;
	}
      }
      closedir(dir);
    }
    else { /* if not a directory, just copy the argv-pointer cstring */
      if(!(name = malloc(strlen(files[file_no]) + 1))) {
	return 0; }
      strncpy(name, files[file_no], strlen(files[file_no]) + 1);
      files[file_no] = name;
    }
  }
  *out_files = files;
  return count;
}

/* Processes each charsave file in 'files'. Returns what read_files() did. */
unsigned int process_files(char **files, const unsigned int file_count) {
  struct charinfo_t **characters;
  struct list_t list;
  char buf[1024] = "";
  unsigned int return_val, char_count = 0;
  if(!(characters = malloc(file_count * sizeof(struct charinfo_t *)))) {
    printf("d2stats: initial malloc failed (not enough memory?)\n");
    return 1; }
  zero_out(characters, file_count * sizeof(struct charinfo_t *));
  zero_out(&list, sizeof(struct list_t));
  return_val = read_files(files, file_count, characters, &list, &char_count);
  config(config_mode, buf);
  if(char_count && !strncmp(buf, "stats", 6)) {
    list_files(characters, &list, file_count); }
  if(list.ear_list) {
    hashtable_destroy(list.ear_list, 1); }
  free_files(characters, files, file_count);
  return return_val;
}
