#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"

/* Extracts the variable-length bytes in a character save file that
   determine the character's statistics - strength, vitality, health, gold,
   etc. Returns the number of bytes extracted. */
unsigned int extract_stat_bytes(FILE *charfile, uint8 **stats) {
  uint8 buffer[60];
  unsigned int len = 0, pos;
  if(read_bytes(charfile, buffer, 0x2fd, 60) != 60) {
    return 0; }
  if(buffer[0] != 'g' || buffer[1] != 'f') {
    return 0; }
  for(pos = 59; pos >= 28; pos--) {
    if(buffer[pos] == 'f' && buffer[pos - 1] == 'i') {
      len = pos - 3;
      break; }
  }
  if(!len || !(*stats = malloc(len))) {
    return 0; }
  memcpy(*stats, buffer + 2, len);
  return len;
}

/* Processes the doubly bit-reversed character stats block at 0x2fd.
   More info about this block can be found at:
   http://www.ladderhall.com/ericjwin/html/d2filestruct.html
   Also reads the 30-byte skill block that comes right afterward, without
   processing it. Returns the length of this block, including header/footer,
   plus the length of the 30-byte skill block. */
unsigned int extract_stats(FILE *charfile, struct stats_t *charstats) {
  uint8 *stats;
  uint16 id;
  unsigned int len, pos;
  if(!(len = extract_stat_bytes(charfile, &stats))) {
    return 0; }
  if(read_bytes(charfile, charstats->skills, 0x2ff + len + 2, 30) != 30) {
    return 0; }
  /* each byte in the block first must be bit-reversed. by that, I mean
     each byte becomes a mirror image of itself. for example,
     '01001101' becomes '10110010'. */
  for(pos = 0; pos < len; pos++) {
    stats[pos] = bitreverse(stats[pos], 8);
  }
  /* roughly, this is the format of the stat block:
     - "gf"
     - ID (9 bits)
     - Value (variable-length, depends on ID)
     - ID
     - Value
     ...
     - Stop ID ('111111111')
     - Zeros to pad the rest of the byte out, if not byte-aligned
     - "if"
     each token is itself bit-reversed, including both the IDs and values,
     and need to be extracted in reverse. */
  for(pos = 9, id = extract_rev_int(stats, 0, 9);
      id != 0x1ff && pos < len * 8;
      id = extract_rev_int(stats, pos, 9), pos += 9) {
    switch(id) {
    case 0x0: /* str - 10 bits */
      charstats->str = extract_rev_int(stats, pos, 10);
      pos += 10;
      break;
    case 0x1: /* ener - 10 bits */
      charstats->ener = extract_rev_int(stats, pos, 10);
      pos += 10;
      break;
    case 0x2: /* dex - 10 bits */
      charstats->dex = extract_rev_int(stats, pos, 10);
      pos += 10;
      break;
    case 0x3: /* vit - 10 bits */
      charstats->vit = extract_rev_int(stats, pos, 10);
      pos += 10;
      break;
    case 0x4: /* stats left - 10 bits */
      charstats->stats_left = extract_rev_int(stats, pos, 10);
      pos += 10;
      break;
    case 0x5: /* skills left - 8 bits */
      charstats->skills_left = extract_rev_int(stats, pos, 8);
      pos += 8;
      break;
    case 0x6: /* current health - 21 bits */
      charstats->hp = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0x7: /* max health - 21 bits */
      charstats->health = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0x8: /* current mana - 21 bits */
      charstats->mp = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0x9: /* max mana - 21 bits */
      charstats->mana = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0xa: /* current stamina - 21 bits */
      charstats->sp = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0xb: /* max stamina - 21 bits */
      charstats->stam = extract_rev_int(stats, pos, 21);
      pos += 21;
      break;
    case 0xc: /* level - 7 bits */
      charstats->level = extract_rev_int(stats, pos, 7);
      pos += 7;
      break;
    case 0xd: /* experience - 32 bits */
      charstats->exp = extract_rev_int(stats, pos, 32);
      pos += 32;
      break;
    case 0xe: /* gold on person - 25 bits */
      charstats->gold = extract_rev_int(stats, pos, 25);
      pos += 25;
      break;
    case 0xf: /* gold in stash - 25 bits */
      charstats->gold_stash = extract_rev_int(stats, pos, 25);
      pos += 25;
      break;
    default: /* unknown - return with error */
      write_log("unknown id = 0x%03x\n", 1, id);
      free(stats);
      return 0;
      break;
    }
  }
  free(stats);
  /* two bytes for 'gf', two bytes for 'if', and 30 bytes for skills */
  return len + 34;
}

/* Extracts and parses the class byte in a charsave file at 0x28.
   Returns the number of bytes extracted. */
unsigned int extract_class(FILE *charfile, struct charinfo_t *charinfo) {
  uint8 buffer;
  if(read_bytes(charfile, &buffer, 0x28, 1) != 1) {
    return 0; }
  switch(buffer) {
  case 0: charinfo->char_class = ama; break;
  case 1: charinfo->char_class = src; break;
  case 2: charinfo->char_class = nec; break;
  case 3: charinfo->char_class = pal; break;
  case 4: charinfo->char_class = bar; break;
  case 5: charinfo->char_class = dru; break;
  case 6: charinfo->char_class = asn; break;
  default: return 0; break;
  }
  return 1;
}

/* Extracts and parses the int at 0x10 that determines which weapon set
   (norm or alt) the character currently is using. Returns number of bytes
   extracted. */
unsigned int extract_weapon_set(FILE *charfile, struct charinfo_t *charinfo) {
  uint8 buffer[4];
  zero_out(buffer, 4);
  if(read_bytes(charfile, buffer, 0x10, 4) != 4) {
    return 0; }
  charinfo->weapon_set = buffer[0] == 1 ? alt_w : norm_w;
  return 4;
}

/* Extracts and stores the cstring at 0x14 that specifies the character's
   name, which can be up to 15 bytes. Returns string length. */
unsigned int extract_name(FILE *charfile, struct charinfo_t *charinfo) {
  unsigned int len;
  uint8 buffer[16];
  zero_out(buffer, 16);
  if(read_bytes(charfile, buffer, 0x14, 15) != 15) {
    return 0; }
  len = buffer[15] != '\0' ? 15 : (unsigned int)strlen((char *)buffer);
  memcpy(charinfo->name, buffer, len);
  buffer[len] = '\0';
  return len;
}

/* Extracts and parses the character title byte at 0x25. Returns number of
   bytes extracted. */
unsigned int extract_title(FILE *charfile, struct charinfo_t *charinfo) {
  uint8 buffer = 0;
  if(read_bytes(charfile, &buffer, 0x25, 1) != 1) {
    return 0; }
  if(buffer >= 5 && buffer < 10) {
    charinfo->char_title = sir; }
  else if(buffer >= 10 && buffer < 15) {
    charinfo->char_title = lord; }
  else if(buffer == 15) {
    charinfo->char_title = baron; }
  else {
    charinfo->char_title = norm_title; }
  return 1;
}

/* Extracts and parses the byte at 0x24 that determines whether the character
   is hardcore, ladder, and has died at least once. Returns number of bytes
   extracted. Returns 0 if character is non-expansion. */
unsigned int extract_chartype(FILE *charfile, struct charinfo_t *charinfo) {
  uint8 buffer = 0;
  if(read_bytes(charfile, &buffer, 0x24, 1) != 1) {
    return 0; }
  if(!(buffer >> 5 & 1)) {
    return 0; }
  charinfo->is_hardcore = (buffer >> 2 & 1) ? yes : no;
  charinfo->is_ladder = (buffer >> 6 & 1) ? yes : no;
  charinfo->has_died = (buffer >> 3 & 1) ? yes : no;
  return 1;
}

/* Extracts and parses quest bytes that specify whether or not the character
   has read the Act 5 scrolls of resistance for normal, nightmare and hell
   modes. Returns the number of bytes extracted. */
unsigned int extract_res(FILE *charfile, struct charinfo_t *charinfo) {
  uint8 buffer = 0;
  if(read_bytes(charfile, &buffer, 0x1a3, 1) != 1) {
    return 0; }
  charinfo->has_res[0] = buffer >> 7 & 1 ? yes : no; 
  if(read_bytes(charfile, &buffer, 0x203, 1) != 1) {
    return 0; }
  charinfo->has_res[1] = buffer >> 7 & 1 ? yes : no; 
  if(read_bytes(charfile, &buffer, 0x263, 1) != 1) {
    return 0; }
  charinfo->has_res[2] = buffer >> 7 & 1 ? yes : no; 
  return 3;
}
