#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "config.h"

/* Config tokens to parse from d2stats.conf - related to the 'config_t' enum */
const unsigned int sizeof_config = 35;
const char *config_names[] = {
  "magic_pre", "magic_suf", "rare_pre", "rare_suf", "armor", "weapons",
  "misc", "unique", "set_item", "rune", "property", "skill", "strtbl",
  "exp_strtbl", "patch_strtbl", "default_mode", "filter_hc", "filter_ladder",
  "season_no", "server_name", "char_xslt", "list_xslt", "charsave",
  "charinfo", "acct", "parse_accts", "report", "parse_reports",
  "report_backup", "report_log", "read_report_log", "data", "cgi_link",
  "xml_char", "xml_list" };

/* cstring representations of command-line arguments. */
const unsigned int sizeof_cmd_args = 4;
const char *cmd_args[] = {
  "-c", "-s", "-l", "-h" };

/* Looks through argv and adjusts the config array 'out' as needed. */
void parse_args(char **out, int argc, char **argv) {
  unsigned int pos, arg;
  for(pos = 1; pos < (unsigned int)argc; pos++) {
    for(arg = 0; arg < sizeof_cmd_args; arg++) {
      if(!strncmp(argv[pos], cmd_args[arg], strlen(cmd_args[arg]) + 1)) {
	/* Be alert, this will likely cause memory leaks since out[param]
	   is likely a malloc'd block. However, it could also be a static
	   string literal, so freeing it would be bad. It's only like 10 bytes
	   and only happens once per run anyway, maybe. */
	switch(arg) {
	case 0: out[config_mode] = "character"; break;
	case 1: out[config_mode] = "stats"; break;
	case 2: out[filter_ladder] = "yes"; break;
	case 3: out[filter_hc] = "yes"; break;
	default: break;
	}
      }
    }
  }
}

/* Parses a config file and makes a cstring array containing configuration
   info (along with defaults if the file is missing entries). Returns null if
   this file can't be opened. */
char **make_config_arr(int argc, char **argv) {
  char **out, buf[1024], name_buf[1024], *name, *val;
  enum config_t param;
  enum flag_t matched;
  FILE *file;
  unsigned int pos, buf_len;
  if(!(out = malloc(sizeof_config * sizeof(char *)))) {
    return NULL; }
  zero_out(out, sizeof_config * sizeof(char *));
  if(!(file = fopen(config_file, "r"))) {
    return NULL; }
  for(fgets(buf, 1024, file); !feof(file); fgets(buf, 1024, file)) {
    /* Ignore blank or '#'-prefixed lines. */
    if(buf[0] != '\0' && buf[0] != '#' && buf[0] != '\n') {
      buf_len = (unsigned int)strlen(buf);
      /* Adjust for fgets newline-at-end-of-string BS. */
      if(buf[buf_len - 1] == '\n') {
	buf_len--; }
      /* Scan for equals, copy everything before it to name, copy everything
         after it to val, trim both. */
      for(pos = 0; pos < buf_len && buf[pos] != '='; pos++);
      strncpy(name_buf, buf, pos);
      name_buf[pos] = '\0';
      name = trim(name_buf);
      if(!(val = malloc(buf_len - pos + 1))) {
	return NULL; }
      strncpy(val, buf + pos + 1, buf_len - pos - 1);
      val[buf_len - pos - 1] = '\0';
      val = trim_realloc(val);
      /* Check to see if name matches anything in the token array. If it
         does, assign the value and break, if it doesn't, warn the user. */
      matched = no;
      for(param = magic_pre_file; param <= xml_list_file; param++) {
	if(!strncmp(name, config_names[param], strlen(config_names[param]) + 1)) {
	  out[param] = val;
	  matched = yes;
	  break; }
      }
      if(matched == no) {
	write_log("unknown param in config: \"%s\"\n", 2, name);
	free(val); }
    }
  }
  /* Set defaults if necessary. */
  for(param = magic_pre_file; param <= xml_list_file; param++) {
    if(!out[param]) {
      out[param] = config_defaults[param]; }
  }
  fclose(file);
  parse_args(out, argc, argv);
  return out;
}

/* Simply sets the config array to defaults, is a fallback in case the
   file can't be read at all. */
char **make_config_default(int argc, char **argv) {
  char **out;
  unsigned int pos;
  if(!(out = malloc(sizeof_config * sizeof(char *)))) {
    write_log("could not allocate space for config data! bailing out\n", 1);
    return NULL; }
  for(pos = 0; pos < sizeof_config; pos++) {
    out[pos] = config_defaults[pos]; }
  parse_args(out, argc, argv);
  return out;
}

/* Wrapper to access the static lookup in config_set. */
void config(enum config_t param, char *buf) {
  config_set(param, buf, 0, NULL);
}

/* Either returns the cstring from the lookup, or reads the config into
   the lookup. Uses param to perform the lookup, writes the cstring to buf. */
void config_set(enum config_t param, char *buf, int argc, char **argv) {
  static char **lookup = NULL;
  if(!lookup) {
    if(!(lookup = make_config_arr(argc, argv))) {
      write_log("failed to read config file, reverting to defaults\n", 2);
      lookup = make_config_default(argc, argv);
    }
  }
  /* the buffer must be pre-allocated, either stack or heap */
  if(buf) {
    /* if it's a file, prefix the data path onto it first */
    if(param <= patch_strtbl_file) {
      strncpy(buf, lookup[data_path], strlen(lookup[data_path]) + 1);
      strncat(buf, DIR_SLASH, 3);
      strncat(buf, lookup[param], strlen(lookup[param]) + 1); }
    else {
      strncpy(buf, lookup[param], strlen(lookup[param]) + 1); }
  }
}

/* Basically all this does is allocate a file cstring array and copies
   everything in argv that isn't a predefined argument to it. Returns
   the amount of filenames detected. */
unsigned int get_files(int argc, char **argv, char ***out) {
  unsigned int pos, out_pos, arg, count = 0;
  enum flag_t match = no;
  if(argc <= 1) {
    return 0; }
  /* Figure out how many filenames there are. */
  for(pos = 1; pos < (unsigned int)argc; pos++) {
    match = no;
    for(arg = 0; arg < sizeof_cmd_args; arg++) {
      if(!strncmp(argv[pos], cmd_args[arg], strlen(cmd_args[arg]) + 1)) {
	match = yes;
	break; }
    }
    if(match == no) {
      count++; }
  }
  if(!count || !(*out = malloc(count * sizeof(char *)))) {
    return 0; }
  /* Pass through again, but now that there's space for the pointers,
     grab the filenames. */
  for(pos = 1, out_pos = 0; pos < (unsigned int)argc && out_pos < count; pos++) {
    match = no;
    for(arg = 0; arg < sizeof_cmd_args; arg++) {
      if(!strncmp(argv[pos], cmd_args[arg], strlen(cmd_args[arg]) + 1)) {
	match = yes;
	break; }
    }
    if(match == no) {
      (*out)[out_pos++] = argv[pos]; }
  }
  return count;
}

/* Determines whether or not a character passes a user-defined filter. For
   now, users can filter only hardcore and ladder characters. Does nothing
   if not in 'stats' mode. */
enum flag_t passes_filter(struct charinfo_t *charinfo) {
  char ladder[1024], hc[1024], mode[1024];
  config(config_mode, mode);
  config(filter_ladder, ladder);
  config(filter_hc, hc);
  if(!strncmp(mode, "character", 10)) {
    return yes; }
  if(!strncmp(ladder, "yes", 4) && charinfo->is_ladder == no) {
    return no; }
  if(!strncmp(hc, "yes", 4) && charinfo->is_hardcore == no) {
    return no; }
  return yes;
}

/* Assigns the filename that should be used for printing 'charinfo' to XML
   character data to 'out'. If none is specified, nothing is printed. */
void get_xml_char_file(struct charinfo_t *charinfo, char *out) {
  char buf[1024], *tok;
  out[0] = '\0';
  config(xml_char_file, buf);
  if(buf[0] == '\0') {
    return; }
  if(!(strpbrk(buf, "%"))) {
    strncpy(out, buf, strlen(buf) + 1);
    return; }
  if(strpbrk(buf, "%") == buf) {
    strncpy(out, charinfo->name, strlen(charinfo->name));
    strncat(out, buf + 1, strlen(buf) - 1);
    return; }
  tok = strtok(buf, "%");
  if(tok != NULL) {
    strncat(out, tok, strlen(tok)); }
  strncat(out, charinfo->name, strlen(charinfo->name));
  tok = strtok(NULL, "%");
  if(tok != NULL) {
    strncat(out, tok, strlen(tok)); }
}

/* Assigns the filename that should be used for printing 'list' to XML
   list data to 'out'. If none is specified, nothing is printed. */
void get_xml_list_file(struct list_t *list, char *out) {
  char buf[1024], buf2[1024], *tok;
  out[0] = '\0';
  config(xml_list_file, buf);
  config(season_no, buf2);
  if(buf[0] == '\0') {
    return; }
  if(!(strpbrk(buf, "%"))) {
    strncpy(out, buf, strlen(buf) + 1);
    return; }
  if(strpbrk(buf, "%") == buf) {
    strncpy(out, buf2, strlen(buf2));
    strncat(out, buf + 1, strlen(buf) - 1);
    return; }
  tok = strtok(buf, "%");
  if(tok != NULL) {
    strncat(out, tok, strlen(tok)); }
  strncat(out, buf2, strlen(buf2));
  tok = strtok(NULL, "%");
  if(tok != NULL) {
    strncat(out, tok, strlen(tok)); }
}
