#ifndef _D2STATS_H__
#define _D2STATS_H__

#include <stdarg.h>
#include <stdio.h>
#include <time.h>

#include "hashtable.h"

typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;

#define sizeof_set 32 /* total number of sets */

#ifdef WIN32
#define DIR_SLASH "\\"
#else
#define DIR_SLASH "/"
#endif

/* Class types. */
enum char_class_t {
  ama, asn, bar, dru, nec, pal, src };

/* Titles. 'Destroyer Name', etc. */
enum char_title_t {
  norm_title, sir, lord, baron };

/* Initially used as merely a way to extract bits from a file, it's now
   being used as a boolean type. */
enum flag_t {
  no, yes };

/* Characters get two different weapon slots, and can alternate between
   them, this enum determines which slots the character's using. */
enum weapon_set_t {
  norm_w, alt_w };

/* These enums are used as indexes when requesting vars from config(). */
enum config_t {
  magic_pre_file, magic_suf_file, rare_pre_file, rare_suf_file,
  armor_file, weapons_file, misc_file, unique_file, set_item_file,
  rune_file, property_file, skill_file, strtbl_file, exp_strtbl_file,
  patch_strtbl_file, config_mode, filter_hc, filter_ladder, season_no,
  server_name, char_xslt, list_xslt, charsave_path, charinfo_path,
  acct_path, parse_accts, report_path, parse_reports, report_backup_path,
  report_log, read_report_log, data_path, cgi_link, xml_char_file,
  xml_list_file };

/* The extracted info from the stats block. */ 
struct stats_t {
  uint16 str, dex, ener, vit;
  uint32 hp, sp, mp;
  uint32 health, stam, mana;
  uint8 level;
  uint32 exp;
  uint8 skills[30], skills_left;
  uint16 stats_left;
  uint32 gold, gold_stash; };

/* The compiled stats from equipped property lists. */
struct bonus_stats_t {
  uint8 str, dex, vit, ener;
  uint32 life, stam, mana;
  uint16 mult_life, mult_mana, mult_ar; /* multipliers */
  short fire, light, cold, pois;
  uint8 max_fire, max_light, max_cold, max_pois;
  uint32 def, ar;
  uint8 block, block_speed;
  uint16 ias, fcr, cb, ds, ow, mf;
  uint8 ll, ml; /* leech */
  uint8 skill0, skill1, skill2, skill3, skills[30]; /* +all, +tab1, ... */
  uint8 high_rune, high_set;
  unsigned int high_rune_count, ears_dropped;
  double ival, rune_val, set_val;
  enum flag_t has_ignore; };

/* Character info structure, holds everything you'd ever want to know about
   a charsave file. */
struct charinfo_t {
  char name[16], *acct, *filename;
  uint8 set_info[sizeof_set];
  unsigned int filesize;
  struct stats_t stats;
  unsigned int stats_len;
  enum char_class_t char_class;
  enum weapon_set_t weapon_set;
  enum char_title_t char_title;
  struct item_t **items;
  struct prop_t *total;
  unsigned int item_count;
  struct bonus_stats_t bonus_stats;
  enum flag_t is_hardcore, is_ladder, has_died, has_res[3]; };

/* Used to store information extracted from an account file. */
struct acct_t {
  unsigned int ip, last_time; };

/* Used to store information extracted from a game report. */
struct game_t {
  char name[64];
  time_t start_time, end_time;
  unsigned int players;
  enum flag_t not_d2; };

/* Used to store top-10 lists, temporary tallies and general server info. */
struct list_t {
  struct charinfo_t *by_exp[50], *by_ival[20], *by_rune[20], *by_gold[20];
  struct charinfo_t *by_mf[20], *by_res[10], *by_skill[10], *by_fcr[10];
  struct charinfo_t *by_ias[10], *by_melee[10], *by_alive[10], *by_set[10];
  struct charinfo_t *by_ears_dropped[10];
  unsigned int char_count, soj_count, annhi_count, torch_count, std_count;
  unsigned int class_count[7], avg_level, avg_class_level[7];
  unsigned int accts, unique_accts, active_accts;
  unsigned int games, games_over_lim, recent, recent_over_lim;
  unsigned int avg_game_len, total_game_len;
  unsigned int recent_avg_len, recent_total_len;
  double avg_game_time[24], recent_avg_time[24];
  double avg_game_time_mod, recent_avg_time_mod;
  double recent_avg_players, avg_players;
  struct hashtable *ear_list;
  char timezone[128]; };

#include "items.h"

/* auxiliary.c */
/* General-purpose functions. */
void write_log(const char *text, const unsigned int log_lvl, ...);
void zero_out(void *bytes, const unsigned int len);
uint32 bitreverse(const uint32 in, const unsigned int len);
uint32 extract_int(const uint8 *bytes, const unsigned int bitpos, const unsigned int bitlen);
uint32 extract_rev_int(const uint8 *bytes, const unsigned int bitpos, const unsigned int bitlen);
unsigned int read_bytes(FILE *charfile, uint8 *bytes, const unsigned int pos, const unsigned int len);
unsigned int get_filesize(FILE *file);
unsigned int find_str(char *big, char *small);
unsigned int count_bits(unsigned int in);
char *trim(char *in);
char *trim_realloc(char *in);
void print_time(const unsigned int in, char *buf);
void print_timeofday(const unsigned int in, char *buf);

/* charstats.c */
/* Basic extraction routines - everything but items. */
unsigned int extract_stat_bytes(FILE *charfile, uint8 **stats);
unsigned int extract_stats(FILE *charfile, struct stats_t *charstats);
unsigned int extract_class(FILE *charfile, struct charinfo_t *charinfo);
unsigned int extract_weapon_set(FILE *charfile, struct charinfo_t *charinfo);
unsigned int extract_name(FILE *charfile, struct charinfo_t *charinfo);
unsigned int extract_title(FILE *charfile, struct charinfo_t *charinfo);
unsigned int extract_chartype(FILE *charfile, struct charinfo_t *charinfo);
unsigned int extract_res(FILE *charfile, struct charinfo_t *charinfo);

/* config.c */
/* Stores, parses and accesses config data. */
void parse_args(char **out, int argc, char **argv);
char **make_config_arr(int argc, char **argv);
char **make_config_default(int argc, char **argv);
void config(enum config_t param, char *buf);
void config_set(enum config_t param, char *buf, int argc, char **argv);
unsigned int get_files(int argc, char **argv, char ***out);
enum flag_t passes_filter(struct charinfo_t *charinfo);
void get_xml_char_file(struct charinfo_t *charinfo, char *out);
void get_xml_list_file(struct list_t *list, char *out);

/* d2stats.c */
/* Contains main. Nothing else here anymore. */

/* charfile.c */
/* What d2stats used to have; a bunch of overwatch functions to control
   extraction, storage, processing and printing of data. */
unsigned int read_file(char *filename, struct charinfo_t **characters, const unsigned int char_no);
void process_file(struct charinfo_t *charinfo, struct list_t *list);
unsigned int read_files(char **files, const unsigned int file_count, struct charinfo_t **characters, struct list_t *list, unsigned int *char_count);
void list_files(struct charinfo_t **characters, struct list_t *list, const unsigned int file_count);
void free_files(struct charinfo_t **characters, char **files, const unsigned int file_count);
unsigned int expand_args(int argc, char **argv, char ***out_files);
unsigned int process_files(char **files, const unsigned int file_count);

/* acctfile.c */
/* Extracts and processes data from account files. */
unsigned int extract_acct(FILE *file, struct acct_t *acct);
unsigned int read_accts(struct acct_t ***accts);
void parse_accounts(struct list_t *list);

/* reptfile.c */
/* Extracts and processes data from game reports. */
time_t get_time_from_cstr(const char *time_buf);
unsigned int extract_report(FILE *file, struct game_t *game);
enum flag_t move_report(const char *file, const char *dir);
unsigned int read_reports(struct game_t ***games);
void tally_game(struct list_t *list, const struct game_t *game, const unsigned int curr_time);
void write_games(struct game_t **games, const unsigned int count);
unsigned int read_games(struct game_t ***games, unsigned int count);
void parse_games(struct list_t *list);

#endif
