#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"

/* Returns the address of the last 'next' entry in a proplist. */
struct prop_t **find_end_of_prop(struct prop_t **in) {
  for(;*in != NULL; in = &(*in)->next);
  return in;
}

/* Copies a proplist to the address at 'dest', returns the address of the last
   'next' entry in this new proplist. */
struct prop_t **prop_deep_copy(struct prop_t **dest, struct prop_t *src) {
  struct prop_t *curr, *tmp, **last = dest;
  for(curr = src; curr != NULL; curr = curr->next) {
    if(!(tmp = malloc(sizeof(struct prop_t)))) {
      return NULL; }
    memcpy(tmp, curr, sizeof(struct prop_t));
    *last = tmp;
    last = &tmp->next;
  }
  return last;
}

/* Determines whether or not an item is currently equipped. Weapons are
   only 'equipped' if their weapon set (norm, alt) matches 'ws'. Charms
   are only 'equipped' if they are located in the player's inventory. */
enum flag_t is_equipped(struct item_t *item, enum weapon_set_t ws) {
  if(item->location == equipped && item->section == player) {
    if(item->worn == right_hand || item->worn == left_hand) {
      return ws == norm_w; }
    else if(item->worn == alt_right_hand || item->worn == alt_left_hand) {
      return ws == alt_w; }
    else {
      return yes; }
  }
  else if(item->location == stored && item->section == player && item->grid_loc == inventory && item->tag[0] == 'c' && item->tag[1] == 'm' && item->tag[2] >= '1' && item->tag[2] <= '3') {
    return yes; }
  return no;
}

/* Deletes the node pointed to by the address in 'prop'. */
void delete_prop(struct prop_t **prop) {
  struct prop_t *tmp = *prop;
  *prop = (*prop)->next;
  free(tmp);
}

/* Extracts an item's 'tag', which determines what type of item it is. */
void get_item_tag(const uint8 *bytes, char *tag) {
  unsigned int pos;
  for(pos = 0; pos < 4; pos++) {
    tag[pos] = extract_int(bytes, pos * 8 + 60, 8); }
  if(tag[3] == ' ') {
    tag[3] = '\0'; }
  tag[4] = '\0';
}

/* Determines how long the raw proplist at 'bytes' + 'bits' is. */
unsigned int get_length_of_proplist(const uint8 *bytes, const unsigned int bytes_size, const unsigned int bits) {
  unsigned int pos, id;
  struct property_entry_t *prop;
  /* Generally, the proplist format is like this:
     - ID (9 bits)
     - Value (any number of bits)
     - Param (might not even be there)
     - ...
     - Terminator ('111111111')
     There isn't much info available on proplists, but the game itself has
     a file in Patch_D2.mpq called 'ItemStatCosts.txt' that includes lots of
     info describing what each prop ID means, how much bits its val/param are,
     how to print them... This file is read by the program and accessible with
     lookup_prop_by_id(). */
  for(pos = bits, id = extract_int(bytes, pos, 9);
      bits < bytes_size * 8 && id != 0x1ff;
      pos += 9, id = extract_int(bytes, pos, 9)) {
    if(!(prop = lookup_prop_by_id(id))) {
      write_log("unknown property %d!\n", 1, id);
      return 0; }
    /* some IDs need some shims. the following IDs always come in min/max or
       min/max/length pairs, and this isn't noted in the file. if the file
       itself doesn't contain entries for the max / length props, the program
       will crash, because the file's invalid. */
    switch(id) {
    case 17: pos += (lookup_prop_by_id(18))->bits; break;
    case 48: pos += (lookup_prop_by_id(49))->bits; break;
    case 50: pos += (lookup_prop_by_id(51))->bits; break;
    case 52: pos += (lookup_prop_by_id(53))->bits; break;
    case 54: pos += (lookup_prop_by_id(55))->bits + (lookup_prop_by_id(56))->bits; break;
    case 57: pos += (lookup_prop_by_id(58))->bits + (lookup_prop_by_id(59))->bits; break;
    default: break;
    }
    pos += prop->param + prop->bits;
  }
  if(id == 0x1ff) {
    return pos + 9 - bits; }
  else {
    write_log("unexpected end of proplist\n", 1);
    return 0; }
}

/* Determines how long an item entry at 'bytes' is. Returns the length.
   More info on the somewhat convoluted D2 item format can be found here:
   http://home.stx.rr.com/svr/formats/fields.htm */
unsigned int get_item_length(const uint8 *bytes, const unsigned int bytes_size) {
  unsigned int bits, affix, pos;
  int setlist[5];
  enum quality_t qual;
  char tag[5];
  struct entry_t *entry;
  zero_out(setlist, 5 * sizeof(int));
  get_item_tag(bytes, tag);
  entry = lookup_item_by_tag(tag);
  if(bytes_size < 12) { /* no item will ever be shorter than 12 bytes */
    write_log("item under 12 bytes!\n", 1);
    return 0; }
  if(bytes[2] & 1) { /* item is ear */
    bits = 70;
    /* look for a null term on the ear name, which is like a cstring with
       seven bits per char rather than eight. */
    for(;extract_int(bytes, bits, 7) != '\0'; bits += 7);
    bits += 7;
    return bits / 8 + 1; }
  if(bytes[2] >> 5 & 1) { /* item is simple (12 bytes) */
    return 12; }
  if(bytes_size < 18) {
    write_log("non-simple item under 18 bytes!\n", 1);
    return 0; }
  bits = 134;
  qual = determine_qual(extract_int(bytes, bits, 4));
  bits += 4;
  if(extract_int(bytes, bits++, 1)) { /* item has alternate graphic field */
    bits += 3; }
  if(extract_int(bytes, bits++, 1)) { /* item has 'class info' field */
    bits += 11; }
  switch(qual) {
  case low_q:
  case high_q: bits += 3; break; /* item has sub-type field */
  case normal_q:
    if(entry) {
      if(entry->type == charm) { /* item has extra charm affix field */
	bits += 12; }
      else if(entry->type == tome) { /* item has spell id field */
	bits += 5; }
      else if(entry->type == body) { /* do these even exist in game... */
	bits += 10; }
    }
    break;
  case magic_q: bits += 22; break; /* item has two affix fields */
  case unique_q:
  case set_q: bits += 12; break; /* item has suid (set/unique id) field */
  case rare_q:
  case crafted_q: /* item has multiple affix fields */
    bits += 16;
    for(affix = 0; affix < 6; affix++) {
      if(extract_int(bytes, bits++, 1)) {
	bits += 11; }
    }
    break;
  default: write_log("item quality = %d\n", 1, qual); return 0;
  }
  if(bytes[3] >> 2 & 1) { /* item has runeword field */
    bits += 16; }
  if(bytes[3] & 1) { /* item has a 7-bit cstring inscription field */
    for(;extract_int(bytes, bits, 7) != '\0'; bits += 7);
    bits += 7; }
  if(entry && entry->type == gold) { /* again, this doesn't exist in game */
    bits += 13; }
  if(extract_int(bytes, bits++, 1)) { /* 'timestamp field', haven't seen it */
    bits += 96; }
  /* to go any further the program absolutely needs the item tag entry and
     prop entries. if they're null, then the user hasn't installed them
     correctly (most likely) and the program must blindly guess at how long
     the remaining part of the item is. */
  if(!entry || !lookup_prop_by_id(0)) {
    /* and to do that, just look for 'JM' for the next item header, 'jf' for
       the merc header, or 'kf' for the merc footer. likely this will also
       trap iron golem items (which have no footer, but appear at the end of
       the file) through the bound. */
    for(pos = bits / 8; pos < bytes_size; pos++) {
      if(pos + 1 < bytes_size &&
	 ((bytes[pos] == 'J' && bytes[pos + 1] == 'M') ||
	  (bytes[pos] == 'j' && bytes[pos + 1] == 'f') ||
	  (bytes[pos] == 'k' && bytes[pos + 1] == 'f'))) {
	return pos; }
    }
    return pos; }
  if(entry->has_def == yes) { /* item has defense field */
    bits += 11; }
  if(entry->no_dura == no || entry->type == weapon) { /* has maxdura field */
    if(extract_int(bytes, bits, 8)) { /* item also has a normal dura field */
      bits += 9; }
    bits += 8; }
  if(entry->has_quantity == yes) { /* item has a quantity field */
    bits += 9; }
  if(bytes[1] >> 3 & 1) { /* item has a socket field */
    bits += 4; }
  if(qual == set_q) { /* item has a set proplist bitfield, with X bits set */
    for(pos = 0; pos < 5; pos++) {
      if(extract_int(bytes, bits++, 1)) {
	setlist[pos] = 1; }
    }
  }
  /* items that make it to this point have a normal proplist, even if it's
     just a terminator */
  bits += get_length_of_proplist(bytes, bytes_size, bits);
  if(qual == set_q) { /* item has X set proplists */
    for(pos = 0; pos < 5; pos++) {
      if(setlist[pos]) {
	bits += get_length_of_proplist(bytes, bytes_size, bits); }
    }
  }
  if(bytes[3] >> 2 & 1) { /* item has a runeword proplist */
    bits += get_length_of_proplist(bytes, bytes_size, bits); }
  /* return the number of bits, and add one if not byte-aligned */
  return bits / 8 + (bits % 8 ? 1 : 0);
}

/* Determines how many items appear in the item section at 'bytes'. Returns
   the length of the item section in bytes, and assigns the number of items
   detected to 'item_count'. */
unsigned int get_item_count(const uint8 *bytes, const unsigned int bytes_size, unsigned int *item_count) {
  unsigned int pos, my_item_count = 0;
  int len = -1; /* I hate using sentinels, but oh well */
  /* Roughly, here's the format for everything after the skill block:
     - "JM"
     - Two-byte short with the total number of items the player has
       (does NOT count gems / runes / jewels that are socketed into other
        items, which still appear in the item section!)
     - Player items (each prefixed with "JM")
     - "JM"
     - Another two-byte short, which is either 0, or:
       - the number of corpses the player has
       - 12 extra bytes
       - "JM"
       - Two-byte short with the total number of items on the player's corpse
       - Corpse items (each prefixed with "JM")
     - "jf"
     If you have a merc (alive or dead):
       - "JM"
       - Two-byte short with the total number of items on the merc
       - Merc items (each prefixed with "JM")
     - "kf"
     - Either a single byte that's either 0, or 1. if 1:
       - A single item that has been turned into an iron golem
         (no "JM" at the front this time)
     - EOF */
  for(pos = 0; pos < bytes_size && len == -1;) {
    if(pos + 1 < bytes_size && bytes[pos] == 'J' && bytes[pos + 1] == 'M') {
      if((pos + 3 < bytes_size /* check for end of section */
	  && bytes[pos + 2] == '\0'
	  && bytes[pos + 3] == '\0'
	  && (bytes[pos + 4] == 'j' || bytes[pos + 4] == 'k')) ||
	 (pos + 21 < bytes_size
	  && bytes[pos + 2] == 1
	  && bytes[pos + 3] == 0
	  && bytes[pos + 16] == 'J'
	  && bytes[pos + 17] == 'M'
	  && bytes[pos + 20] == 'J'
	  && bytes[pos + 21] == 'M')) {
	len = pos;
      }
      else { /* otherwise, it's just the JM header for a normal item */
	pos += 2;
	my_item_count++; }
    }
    else if(pos + 1 < bytes_size && /* check for end of section without a JM */
	    ((bytes[pos] == 'j' && bytes[pos + 1] == 'f') ||
	     (bytes[pos] == 'k' && bytes[pos + 1] == 'f'))) {
      len = pos;
    }
    else { /* if JM and jf/kf aren't present, it's an item, check its length */
      pos += get_item_length(bytes + pos, bytes_size - pos);
      if(pos == bytes_size) { /* check for end of block (iron golem) */
	my_item_count++;
	len = pos; }
    }
  }
  *item_count = my_item_count;
  return len < 0 ? 0 : len;
}

/* Allocates blocks for items and copies bytes for each item found at 'bytes'
   into each block. Returns the number of items found. */
unsigned int extract_items(const uint8 *bytes, const unsigned int bytes_size, struct item_t **items) {
  unsigned int pos, item_no = 0, item_size = 0;
  for(pos = 0; pos < bytes_size; item_no++) {
    if(!(items[item_no] = malloc(sizeof(struct item_t)))) {
      return 0; }
    zero_out(items[item_no], sizeof(struct item_t));
    if(bytes[pos] == 'J' && bytes[pos + 1] == 'M') { /* adjust for header */
      pos += 2; }
    item_size = get_item_length(bytes + pos, bytes_size - pos);
    items[item_no]->size = item_size;
    pos += item_size;
    if(!(items[item_no]->bytes = malloc(item_size))) {
      return 0; }
    memcpy(items[item_no]->bytes, bytes + (pos - item_size), item_size);
  }
  return item_no;
}

/* Extracts item blocks for an entire item section at 'bytes'. Returns the
   length of the item section. */
unsigned int extract_item_section(const uint8 *bytes, const unsigned int bytes_size, struct charinfo_t *charinfo) {
  unsigned int len, item_count;
  len = get_item_count(bytes, bytes_size, &item_count);
  if(item_count) {
    charinfo->item_count += item_count;
    /* if already malloc'd, realloc instead. */
    if(!(charinfo->items = charinfo->items ? realloc(charinfo->items, charinfo->item_count * sizeof(struct item_t *)) : malloc(charinfo->item_count * sizeof(struct item_t *)))) {
      return 0; }
    if(item_count != extract_items(bytes, len, charinfo->items + (charinfo->item_count - item_count))) {
      return 0; }
  }
  return len;
}

/* Marks each item structure from 'charinfo->items[start]' to
   'charinfo->items[end]' as belonging to section 'section'. */
void mark_item_section(const unsigned int start, const unsigned int end, enum item_section_t section, struct charinfo_t *charinfo) {
  unsigned int item_no;
  for(item_no = start; item_no < end; item_no++) {
    charinfo->items[item_no]->section = section;
  }
}

/* Extracts item blocks for each item section in a character file. Also
   calls item analysis functions to fill the rest of the structure. Returns
   number of items extracted. */
unsigned int extract_item_sections(FILE *charfile, struct charinfo_t *charinfo) {
  unsigned int buffer_size, pos, section_len, item_count = 0;
  uint8 *buffer;
  buffer_size = charinfo->filesize - (0x2fd + charinfo->stats_len);
  if(!(buffer = malloc(buffer_size))) {
    return 0; }
  if(read_bytes(charfile, buffer, 0x2fd + charinfo->stats_len, buffer_size) != buffer_size) {
    return 0; }
  /* Read comments in get_item_count() if you want to know the format for
     the item sections, which will help explain what's going on here. */
  if(buffer[0] != 'J' || buffer[1] != 'M') {
    return 0; }
  pos = 4;
  /* player section */
  section_len = extract_item_section(buffer + pos, buffer_size - pos, charinfo);
  pos += section_len;
  mark_item_section(item_count, charinfo->item_count, player, charinfo);
  item_count = charinfo->item_count;
  if(buffer[pos] != 'J' || buffer[pos + 1] != 'M') {
    return 0; }
  if(buffer[pos + 2] == 1) {
    /* corpse section */
    pos += 16;
    if(buffer[pos] != 'J' || buffer[pos + 1] != 'M') {
      return 0; }
    pos += 4;
    section_len = extract_item_section(buffer + pos, buffer_size - pos, charinfo);
    pos += section_len;
    mark_item_section(item_count, charinfo->item_count, corpse, charinfo);
    item_count = charinfo->item_count;
  }
  else {
    pos += 4; }
  if(buffer[pos] == 'j' && buffer[pos + 1] == 'f') {
    /* merc section */
    pos += 2;
    if(buffer[pos] == 'J' && buffer[pos + 1] == 'M') {
      pos += 4;
      section_len = extract_item_section(buffer + pos, buffer_size - pos, charinfo);
      pos += section_len;
      mark_item_section(item_count, charinfo->item_count, merc, charinfo);
      item_count = charinfo->item_count;
    }
    if(buffer[pos] != 'k' || buffer[pos + 1] != 'f') {
      return 0; }
    pos += 2;
    if(buffer[pos] == 1) { /* iron golem */
      pos += 3;
      section_len = extract_item_section(buffer + pos, buffer_size - pos, charinfo);
      pos += section_len;
      charinfo->items[item_count]->section = ig;
    }
  }
  free(buffer);
  analyze_items(charinfo);
  return item_count;
}

/* Frees item blocks, item proplists and items themselves in 'charinfo'. */
void free_items(struct charinfo_t *charinfo) {
  unsigned int item_no, pos;
  struct prop_t *prop, *tmp;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    for(prop = charinfo->items[item_no]->norm_pl; prop != NULL;) {
      tmp = prop->next;
      free(prop);
      prop = tmp;
    }
    for(pos = 0; pos < 5; pos++) {
      for(prop = charinfo->items[item_no]->set_pl[pos]; prop != NULL;) {
	tmp = prop->next;
	free(prop);
	prop = tmp;
      }
    }
    free(charinfo->items[item_no]->bytes);
    free(charinfo->items[item_no]);
  }
  if(charinfo->item_count) {
    free(charinfo->items); }
}
