#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"
#include "tally.h"

/* Returns the quality specified by raw extracted val 'qual'. */
enum quality_t determine_qual(const uint8 qual) {
  switch(qual) {
  case 1: return low_q;
  case 2: return normal_q;
  case 3: return high_q;
  case 4: return magic_q;
  case 5: return set_q;
  case 6: return rare_q;
  case 7: return unique_q;
  case 8: return crafted_q;
  default: return unknown_q;
  }
}

/* Calculates bonuses granted by item property lists such as damage, defense,
   durability and block rating. */
void analyze_extended_props(struct item_t *item, const uint8 *set_info, const enum char_class_t char_class) {
  struct entry_t *entry;
  unsigned int add_min = 0, add_max = 0, add_ac = 0, add_block = 0, add_dura = 0;
  double mult_min = 1.0, mult_max = 1.0, mult_ac = 1.0, mult_dura = 1.0, mult_req = 1.0;
  if(!(entry = lookup_item_by_tag(item->tag))) {
    return; }
  if(item->item_class == weapon_c) {
    item->dmg_enhanced = get_dmg(item, &mult_min, &mult_max, &add_min, &add_max, set_info);
    item->min_dmg = entry->min_dmg;
    item->max_dmg = entry->max_dmg;
    if(item->is_ethereal) { /* ethereal weapons get 150% dmg */
      item->min_dmg = (uint16)(item->min_dmg * 1.5);
      item->max_dmg = (uint16)(item->max_dmg * 1.5); }
    item->min_dmg = (uint16)(item->min_dmg * mult_min);
    item->max_dmg = (uint16)(item->max_dmg * mult_max);
    item->min_dmg += add_min;
    item->max_dmg += add_max;
  }
  if(item->def) {
    item->ac_enhanced = get_def(item, &mult_ac, &add_ac, set_info);
    item->ac = item->def;
    item->ac = (uint16)(item->ac * mult_ac);
    item->ac += add_ac;
  }
  if(item->item_class == shield_c) {
    item->block_enhanced = get_block(item, &add_block, set_info);
    item->block = entry->block + block_by_class(char_class);
    item->block += add_block;
    if(item->block > 75) { /* shields can only have up to 75% block */
      item->block = 75; }
  }
  if(item->max_dura) {
    item->dura_enhanced = get_dura(item, &mult_dura, &add_dura);
    if(mult_dura != 0.0) { /* indestructible items have no dura */
      item->mod_dura = item->dura;
      item->mod_max_dura = (uint16)(item->max_dura * mult_dura) + add_dura;
    }
    else {
      item->indestructible = yes; }
  }
  if(entry->req_level || entry->req_str || entry->req_dex) {
    item->req_enhanced = get_req(item, &mult_req);
    /* The usual item tables store their own required level, but set and
       unique items have their own table with their own required level
       entries. Level requirements aren't affected by 'Requirements -X%' */
    if(entry->req_level) {
      item->req_lvl = entry->req_level;
      if(item->qual == set_q && lookup_set_item_by_id(item->suid)) {
	item->req_lvl = lookup_set_item_by_id(item->suid)->req_level; }
      if(item->qual == unique_q && lookup_unique_by_id(item->suid)) {
	item->req_lvl = lookup_unique_by_id(item->suid)->req_level; }
    }
    if(entry->req_str) {
      item->req_str = (int)floor((entry->req_str * mult_req) + 0.5);
    }
    if(entry->req_dex) {
      item->req_dex = (int)floor((entry->req_dex * mult_req) + 0.5);
    }
  }
}

/* Adds the property value / param at 'bytes' + 'bits' with id 'id' to the
   tail of the proplist, 'prop'. Returns the bit position for the next id. */
unsigned int add_prop(const uint8 *bytes, const unsigned int bits, struct prop_t **prop, const unsigned int id) {
  struct property_entry_t *entry, *entry2, *entry3;
  unsigned int pos = bits;
  if(!(*prop = malloc(sizeof(struct prop_t)))) {
    return 0; }
  zero_out(*prop, sizeof(struct prop_t));
  if(!(entry = lookup_prop_by_id(id))) {
    free(*prop);
    return 0; }
  (*prop)->id = id;
  if(entry->param) { /* the param might not exist at all */
    (*prop)->param = extract_int(bytes, pos, entry->param);
    pos += entry->param; }
  (*prop)->val = extract_int(bytes, pos, entry->bits) - entry->bias;
  pos += entry->bits;
  /* Depending on the 'encode' entry in the ItemStatCost.txt table,
     the param/val pair might need to be decoded into meaningful values.
     Examples of properties that use these encodings are provided. */
  switch(entry->encode) {
  case 1: /* '+2 to Magic Arrow (Amazon Only)' */
    (*prop)->skill_id = (*prop)->param;
    (*prop)->skill_level = (*prop)->val;
    break;
  case 2: /* '10% chance to cast Level 2 Charged Bolt on striking' */
    (*prop)->skill_id = (*prop)->param >> 6 & 0x3ff;
    (*prop)->skill_level = (*prop)->param & 0x3f;
    break;
  case 3: /* 'Level 1 Teleport (50/50 Charges)' */
    (*prop)->skill_id = (*prop)->param >> 6 & 0x3ff;
    (*prop)->skill_level = (*prop)->param & 0x3f;
    (*prop)->charges = (*prop)->val & 0xff;
    (*prop)->max_charges = (*prop)->val >> 8 & 0xff;
    break;
  case 4: /* '+10 Strength (Increases Near Dusk)' */
    (*prop)->time = (*prop)->val >> 20 & 0x3;
    (*prop)->val_during_time = (*prop)->val & 0x3ff;
    (*prop)->val_off_time = (*prop)->val >> 10 & 0x3ff;
    break;
  default: break;
  }
  /* Again, the following properties come in min/max or min/max/length pairs
     and need to be shimmed properly. */
  switch(id) {
  case 17:
  case 48:
  case 50:
  case 52:
    (*prop)->min = (*prop)->val;
    if(!(entry2 = lookup_prop_by_id(id + 1))) {
      free(*prop);
      return 0; }
    (*prop)->max = extract_int(bytes, pos, entry2->bits) - entry2->bias;
    pos += entry2->bits;
    break;
  case 54:
  case 57:
    (*prop)->min = (*prop)->val;
    if(!(entry2 = lookup_prop_by_id(id + 1))) {
      free(*prop);
      return 0; }
    (*prop)->max = extract_int(bytes, pos, entry2->bits) - entry2->bias;
    pos += entry2->bits;
    if(!(entry3 = lookup_prop_by_id(id + 2))) {
      free(*prop);
      return 0; }
    (*prop)->length = extract_int(bytes, pos, entry3->bits) - entry3->bias;
    pos += entry3->bits;
    break;
  default: break;
  }
  return pos - bits;
}
 
/* Compacts a proplist so that any duplicate properties are totaled and
   combined into one property. '+5 to Strength, +5 to Strength' becomes
   '+10 to Strength'. */
void compile_props(struct prop_t **prop) {
  struct prop_t **curr, **copy;
  for(curr = prop; *curr != NULL; curr = &(*curr)->next) {
    for(copy = &(*curr)->next; *copy != NULL;) {
      /* check to see whether copy has the same id as curr.
         an example of id 188 is '+3 to Warcries (Barbarian Only)' and
         the param needs to be checked to make sure both add to the same
         class / skilltab, so that '+3 to Warcries' and '+3 to Masteries'
         are not combined into '+6 to Warcries'. */
      if(((*copy)->id == (*curr)->id &&
	  (*copy)->skill_id == (*curr)->skill_id &&
	  (*copy)->time == (*curr)->time) &&
	 ((*copy)->id != 188 || (*copy)->param == (*curr)->param)) {
	(*curr)->val += (*copy)->val;
	(*curr)->skill_level += (*copy)->skill_level;
	(*curr)->max_charges += (*copy)->max_charges;
	(*curr)->charges += (*copy)->charges;
	(*curr)->val_during_time += (*copy)->val_during_time;
	(*curr)->val_off_time += (*copy)->val_off_time;
	(*curr)->min += (*copy)->min;
	(*curr)->max += (*copy)->max;
	/* an example of id 54 is 'Adds 5-10 Cold Damage'. the cold length
	   (which is hidden) stacks, poison length doesn't stack like this. */
	if((*curr)->id == 54) {
	  (*curr)->length += (*copy)->length; }
	delete_prop(copy);
      }
      /* the following IDs are for extra min/max one-handed or
	 two-handed damage. if curr is one-handed and the copy is two-handed,
	 remove the redundant two-handed copy. */
      else if((*copy)->id == 23 && (*curr)->id == 21) {
	delete_prop(copy); }
      else if((*copy)->id == 24 && (*curr)->id == 22) {
	delete_prop(copy); }
      /* 159 and 160 are for extra min/max throw damage, just remove them. */
      else if((*copy)->id == 159 || (*copy)->id == 160) {
	delete_prop(copy); }
      else {
	copy = &(*copy)->next; }
    }
    if(*curr == NULL) {
      break; }
  }
}

/* Reads the property list at 'bytes' + 'bits' into 'prop'. Returns the number
   of bits read. */
unsigned int read_proplist(const uint8 *bytes, struct prop_t **prop, const unsigned int bits) {
  unsigned int pos, id, count;
  struct prop_t **curr = prop;
  /* Read comments at get_length_of_proplist() if you want to know the format
     for property lists so you can understand what's going on here. */
  for(pos = bits, id = extract_int(bytes, pos, 9), count = 0;
      id != 0x1ff;
      pos += 9, count++, id = extract_int(bytes, pos, 9)) {
    if(!(lookup_prop_by_id(id))) {
      write_log("unknown property %d!\n", 1, id);
      return 0; }
    pos += add_prop(bytes, pos + 9, curr, id);
    curr = &(*curr)->next;
  }
  if(id == 0x1ff) {
    return pos - bits; }
  else {
    write_log("unexpected end of proplist\n", 1);
    return 0; }
}

/* Analyzes a non-simple (over 12 bytes) item, filling its item structure 
   with useful information about the item.
   Info on the Diablo II item format can be found here:
   http://home.stx.rr.com/svr/formats/fields.htm */
void analyze_extended_item(struct item_t *item) {
  unsigned int bits, affix, pos;
  struct entry_t *entry;
  struct prop_t **tail;
  char *tag = item->tag;
  entry = lookup_item_by_tag(tag);
  /* if the tables aren't loaded, this function can't do anything. */
  if(!entry || !lookup_prop_by_id(0)) {
    return; }
  bits = 95;
  /* the GUID is an item's unique 32-bit identifier, if two items have the
     same GUID in the same game, they're dupes and one is automatically
     deleted by Diablo II. */
  item->guid = extract_int(item->bytes, bits, 32);
  bits += 32;
  /* an item's 'item level', commonly referred to as ilvl, is one quantity used
     by the game to determine a lot of its properties on drop. */
  item->ilvl = extract_int(item->bytes, bits, 7);
  bits += 7;
  /* normal, magic ('blue'), rare ('yellow'), unique ('gold'), set ('green'),
     crafted ('orange'), etc. */
  item->qual = determine_qual(extract_int(item->bytes, bits, 4));
  bits += 4;
  /* rings, charms and amulets can use any of around 7 different values here
     to determine what it looks like. other item types usually have graphics
     determined from table values */
  if((item->has_alt_gfx = extract_int(item->bytes, bits++, 1) ? yes : no)) {
    item->alt_gfx = extract_int(item->bytes, bits, 3);
    bits += 3; }
  /* honestly, I have no clue what the hell this is supposed to be */
  if((item->has_class_info = extract_int(item->bytes, bits++, 1) ? yes : no)) {
    item->class_info = extract_int(item->bytes, bits, 11);
    bits += 11; }
  switch(item->qual) {
  case low_q:
  case high_q:
    /* 'cracked', 'broken', 'low quality', 'high quality', etc. */
    item->subqual = extract_int(item->bytes, bits, 3);
    bits += 3;
    break;
  case normal_q:
    if(entry) {
      /* besides tomes, I don't think this code is used. */
      if(entry->type == charm) {
	/* apparently charms can have their own affix/suffix. */
	item->charm_suff = extract_int(item->bytes, bits++, 1) ? yes : no;
	item->charm_affix = extract_int(item->bytes, bits, 11) ? yes : no;
	bits += 11; }
      else if(entry->type == tome) {
	/* this field determines what spell the tome / scroll casts. however,
	   the only tomes that exist in the game cast either 'town portal'
	   or 'identify', and this is clearly noted by the tome name anyway */
	bits += 5; }
      else if(entry->type == body) {
	/* this field apparently is used to determine what monster this body
	   part came from. body parts do not exist in Diablo II, my guess is
	   that this was part of a necro spell that never made it in */
	bits += 10; }
    }
    break;
  case magic_q:
    /* affix IDs can be indexed in a table to determine an item's name.
       Diablo II also uses affixes to determine an item's proplist.
       for example, a blue item might be called 'Godly Plate of the Whale' */
    item->affix_count = 0;
    /* 'Godly' */
    if((item->affix[0] = extract_int(item->bytes, bits, 11))) {
      item->has_aff[0] = yes;
      item->affix_count++; }
    bits += 11;
    /* 'of the Whale' */
    if((item->affix[1] = extract_int(item->bytes, bits, 11))) {
      item->has_aff[1] = yes;
      item->affix_count++; }
    bits += 11;
    break;
  case set_q: 
  case unique_q:
    /* the item's suid (set/unique id) determines what type of set item or
       unique it is. this id is indexed in various tables later.
       yes, I know 'suid' also means something else in Unix, deal with it */
    item->suid = extract_int(item->bytes, bits, 12);
    bits += 12;
    break;
  case rare_q:
  case crafted_q:
    /* each rare or crafted item has two rare names and up to six affixes.
       the ID is indexed in a rare name table later. for example, a yellow bow
       might be randomly assigned the name 'Grim Nock' */
    /* 'Grim' */
    item->name1 = extract_int(item->bytes, bits, 8);
    bits += 8;
    /* 'Nock' */
    item->name2 = extract_int(item->bytes, bits, 8);
    bits += 8;
    /* these affixes are never displayed, but store them anyway */
    for(affix = 0; affix < 6; affix++) {
      if((item->has_aff[affix] = extract_int(item->bytes, bits++, 1) ? yes : no)) {
	item->affix[affix] = extract_int(item->bytes, bits, 11);
	bits += 11; }
    }
    break;
  case 9: write_log("item quality = 9\n", 2); bits += 16; break; /* ? */
  default: write_log("item quality = %d\n", 1, item->qual); return;
  }
  /* items with runewords embedded have their own runeword ID that has to be
     looked up in a table. I'm not entirely clear on the format for this ID,
     it's something like, subtract 25 if less than 0x80, or 26 if greater, 
     then check the table. no there isn't a dummy entry in the table anywhere
     that would cause this inconsistency, I'm not sure why it's that way. */
  if(item->bytes[3] >> 2 & 1) {
    item->runeword = extract_int(item->bytes, bits, 16) & 0xfff;
    bits += 16; }
  /* inscriptions and ear names are stored as 7-bit cstrings. */
  if(item->bytes[3] & 1) {
    for(pos = 0;
	pos < 16 && (item->inscription[pos] = (extract_int(item->bytes, bits, 7))) != '\0';
	pos++, bits += 7);
    item->inscription[pos] = '\0';
    bits += 7; }
  /* in misc.txt there's a 'gold' item type that you store in your inventory
     like you did in Diablo 1. I guess that was dropped, this field notes how
     much gold there'd be in such an item */
  if(entry && entry->type == gold) {
    bits += 13; }
  /* 'timestamp'... what the hell, I've yet to see this */
  if(extract_int(item->bytes, bits++, 1)) {
    bits += 96; }
  /* armor items have their own defense field. this is adjusted for ethereal
     / etc. but isn't adjusted for property list bonuses, such as
     '+200% Enhanced Defense' */
  if(entry->has_def == yes) {
    item->def = extract_int(item->bytes, bits, 11) - 10;
    bits += 11; }
  /* most equippable non-rings have at least a max durability field. if the
     item has 'Indestructible' in its property list, this value is zero,
     otherwise there's also a current durability field. */
  if(entry->no_dura == no || entry->type == weapon) {
    item->max_dura = extract_int(item->bytes, bits, 8);
    bits += 8;
    if(item->max_dura) {
      item->dura = extract_int(item->bytes, bits, 9);
      bits += 9; }
  }
  /* arrows, javelins, throwing knives, tomes, etc. etc. have a quantity
     field that determine how many items are left. */
  if(entry->has_quantity == yes) {
    item->quantity = extract_int(item->bytes, bits, 9);
    bits += 9; }
  /* socketed items always have a field here that tell you how many
     total sockets there are. */
  if(item->bytes[1] >> 3 & 1) {
    item->sockets = extract_int(item->bytes, bits, 4);
    bits += 4; }
  /* set items have a bitfield that determine whether or not they have
     property lists for set bonuses. they can have up to 5, and the
     positioning of its corresponding bit here determines how many set items
     you need to wear to activate that bonus. for example, if bit 2 were
     set here, there'd be one proplist after the normal one, and to activate
     the bonuses in that proplist you'd need to wear at least 4 items from
     that set. the last bit always means 'full set'. */
  if(item->qual == set_q) {
    for(pos = 0; pos < 5; pos++) {
      if(extract_int(item->bytes, bits++, 1)) {
	item->has_set_pl[pos] = yes; }
    }
  }
  /* normal proplist */
  bits += read_proplist(item->bytes, &item->norm_pl, bits);
  /* set proplists */
  if(item->qual == set_q) {
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos]) {
	bits += 9;
	bits += read_proplist(item->bytes, &item->set_pl[pos], bits); }
    }
  }
  /* runeword proplist - this used to be stored separately, but is now
     simply attached to the normal proplist */
  if(item->bytes[3] >> 2 & 1) {
    bits += 9;
    for(tail = &item->norm_pl; *tail != NULL; tail = &(*tail)->next);
    bits += read_proplist(item->bytes, tail, bits); }
  /* this is kind of a weird shim: wands, mace-class weapons and staves
     all have an inherent 50% bonus damage against undead, but this is
     never reported anywhere, so the property is merely added manually here. */
  item->item_class = get_class(item);
  if(!strncmp(entry->type_tag, "wand", 5) ||
     !strncmp(entry->type_tag, "club", 5) ||
     !strncmp(entry->type_tag, "scep", 5) ||
     !strncmp(entry->type_tag, "mace", 5) ||
     !strncmp(entry->type_tag, "staf", 5)) {
    for(tail = &item->norm_pl; *tail != NULL; tail = &(*tail)->next);
    if(!(*tail = malloc(sizeof(struct prop_t)))) {
      return; }
    zero_out(*tail, sizeof(struct prop_t));
    (*tail)->id = 122;
    (*tail)->val = 50;
  }
}

/* Returns the base amount of block value that a class has when using a
   shield with zero block rating. */
unsigned int block_by_class(const enum char_class_t char_class) {
  switch(char_class) {
  case pal: return 30;
  case ama:
  case asn:
  case bar: return 25;
  case dru:
  case nec:
  case src: return 20;
  default: return 0;
  }
}

/* Using the item location bitfield, determine where the item is located. */
void determine_item_loc(struct item_t *item) {
  /* 'location' refers to where you'd see the item on the game screen. */
  switch(extract_int(item->bytes, 42, 3)) {
  case  0: item->location = stored; break;
  case  1: item->location = equipped; break;
  case  2: item->location = belt; break;
  case  4: item->location = moved; break;
  case  6: item->location = socketed; break;
  default: item->location = unknown_loc; break;
  }
  /* 'worn' is only applicable if the character has equipped the item. */
  switch(extract_int(item->bytes, 45, 4)) {
  case  1: item->worn = head; break;
  case  2: item->worn = neck; break;
  case  3: item->worn = chest; break;
  case  4: item->worn = right_hand; break;
  case  5: item->worn = left_hand; break;
  case  6: item->worn = right_finger; break;
  case  7: item->worn = left_finger; break;
  case  8: item->worn = waist; break;
  case  9: item->worn = feet; break;
  case 10: item->worn = hands; break;
  case 11: item->worn = alt_right_hand; break;
  case 12: item->worn = alt_left_hand; break;
  default: item->worn = not_worn; break;
  }
  /* 'grid loc' is only applicable if the item is stored somewhere.*/
  switch(extract_int(item->bytes, 57, 3)) {
  case  1: item->grid_loc = inventory; break;
  case  4: item->grid_loc = cube; break;
  case  5: item->grid_loc = stash; break;
  default: item->grid_loc = not_stored; break;
  }
}

/* Sets various flags in the item structure depending on whether certain bits
   in the item block have been set. */
void set_item_flags(struct item_t *item) {
  item->is_identified = item->bytes[0] >> 4 & 1 ? yes : no;
  item->is_disabled = item->bytes[1] & 1 ? yes : no; /* ? */
  item->is_socketed = item->bytes[1] >> 3 & 1 ? yes : no;
  item->is_illegal = item->bytes[1] >> 6 & 1 ? yes : no; /* ? */
  item->is_ear = item->bytes[2] & 1 ? yes : no;
  item->is_newbie = item->bytes[2] >> 1 & 1 ? yes : no; /* sells for 1g */
  item->is_simple = item->bytes[2] >> 5 & 1 ? yes : no; /* simple = 12 bytes */
  item->is_ethereal = item->bytes[2] >> 6 & 1 ? yes : no;
  item->is_personalized = item->bytes[3] & 1 ? yes : no;
  item->has_runeword = item->bytes[3] >> 2 & 1 ? yes : no;
}

/* Extracts ear information from an item with 'is_ear' == yes. */
void get_ear_info(struct item_t *item) {
  unsigned int pos;
  struct ear_t *ear = &item->ear_info;
  switch(extract_int(item->bytes, 60, 3)) {
  case 0: ear->ear_class = ama; break;
  case 1: ear->ear_class = src; break;
  case 2: ear->ear_class = nec; break;
  case 3: ear->ear_class = pal; break;
  case 4: ear->ear_class = bar; break;
  case 5: ear->ear_class = dru; break;
  case 6: ear->ear_class = asn; break;
  }
  ear->level = extract_int(item->bytes, 63, 7);
  for(pos = 0; pos < 16; pos++) {
    /* 7-bit characters, etc. etc. */
    ear->name[pos] = extract_int(item->bytes, pos * 7 + 70, 7);
    if(ear->name[pos] == '\0') {
      break; }
  }
  for(;pos < 16; pos++) {
    ear->name[pos] = '\0';
  }
}

/* Copies the proplist of a socketed gem, rune or jewel to the end of the
   normal property list of its container item. */
void get_socket_prop(struct item_t *item) {
  struct entry_t *entry;
  struct prop_t **curr;
  if(!(entry = lookup_item_by_tag(item->tag))) {
    return; }
  for(curr = &item->contained_in->norm_pl; *curr != NULL; curr = &(*curr)->next);
  if(!strncmp(entry->type_tag, "gem", 3)) {
    get_gem_prop(curr, item->tag, item->contained_in->item_class); }
  else if(!strncmp(entry->type_tag, "rune", 5)) {
    get_rune_prop(curr, item->tag, item->contained_in->item_class); }
  else {
    curr = prop_deep_copy(curr, item->norm_pl);
  }
}

/* Determine which set items are currently equipped by the character. */
void scan_sets(struct charinfo_t *charinfo) {
  unsigned int item_no, set_id, set_item_id;
  struct item_t *item;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    item = charinfo->items[item_no];
    if(item->qual == set_q && is_equipped(item, charinfo->weapon_set) == yes) {
      set_id = lookup_set_id_by_suid(item->suid);
      set_item_id = lookup_set_item_id_by_suid(item->suid);
      charinfo->set_info[set_id] |= 1 << set_item_id;
    }
  }
}

/* Analyzes items, filling each item structure with useful information
   about the item. Also totals various equipped bonuses before the items are
   freed. Info on the Diablo II item format can be found here:
   http://home.stx.rr.com/svr/formats/fields.htm */
void analyze_items(struct charinfo_t *charinfo) {
  unsigned int item_no, socket_count = 0;
  struct item_t *item, *socket_item;
  struct entry_t *entry;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    item = charinfo->items[item_no];
    item->id = item_no;
    determine_item_loc(item);
    set_item_flags(item);
    if(item->is_ear) {
      get_ear_info(item); }
    else {
      get_item_tag(item->bytes, item->tag);
      if((entry = lookup_item_by_tag(item->tag))) {
	item->name = entry->name; }
      /* Every item has a field that determines how many gems, runes or jewels
	 have been socketed. If this field is nonzero, the item will ALWAYS
	 be followed by the gems, runes or jewels that are contained. */
      item->sockets_used = extract_int(item->bytes, 92, 3);
      /* set the socket count if this item contains sockets */
      if(item->sockets_used) {
	socket_item = item;
	socket_count = item->sockets_used; }
      /* if a socket count is in effect, this item is socketed into another
	 item - figure out which and set item pointers appropriately, then
         decrement the socket count */
      else if(socket_count) {
	socket_item->socketed_items[socket_item->sockets_used - socket_count] = item;
	item->contained_in = socket_item;
	socket_count--; }
      if(item->is_simple == no) {
	analyze_extended_item(item); }
      if(item->contained_in) {
	get_socket_prop(item); }
    }
    /* if stored, these fields determine coordinates of the item's storage
       space. if not, these will be meaningless values */
    item->col = extract_int(item->bytes, 49, 4);
    item->row = extract_int(item->bytes, 53, 4);
    analyze_extended_props(item, charinfo->set_info, charinfo->char_class);
  }
  /* calculate equipped bonuses for the character - this should really be
     its own function */
  scan_sets(charinfo);
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    compile_props(&charinfo->items[item_no]->norm_pl); }
  charinfo->total = generate_total_props(charinfo);
  tally_props(&charinfo->total, charinfo);
  calc_ac(charinfo);
  calc_block(charinfo);
  calc_ival(charinfo);
  calc_high_set(charinfo);
}
