#include <stdio.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"
#include "print.h"

/* Converts an item_location_t enum to a cstr. */
char *item_location_to_cstr(const enum item_location_t location) {
  switch(location) {
  case stored: return "stored";
  case equipped: return "equipped";
  case belt: return "in belt";
  case moved: return "being moved";
  case socketed: return "socketed into another item";
  default: return "unknown location";
  }
}

/* Converts an item_worn_t enum to a cstr. */
char *item_worn_to_cstr(const enum item_worn_t worn) {
  switch(worn) {
  case head: return "head";
  case neck: return "neck";
  case chest: return "chest";
  case right_hand: return "right hand";
  case left_hand: return "left hand";
  case right_finger: return "right finger";
  case left_finger: return "left finger";
  case waist: return "waist";
  case feet: return "feet";
  case hands: return "hands";
  case alt_right_hand: return "alternate right hand";
  case alt_left_hand: return "alternate left hand";
  default: return "not equipped";
  }
}

/* Converts an item_section_t enum to a cstr. */
char *item_section_to_cstr(const enum item_section_t section) {
  switch(section) {
  case player: return "player";
  case corpse: return "corpse";
  case merc: return "mercenary";
  case ig: return "iron golem";
  default: return "unknown";
  }
}

/* Converts an item_grid_loc_t enum to a cstr. */
char *item_grid_loc_to_cstr(const enum item_grid_loc_t grid) {
  switch(grid) {
  case inventory: return "inventory";
  case cube: return "cube";
  case stash: return "stash";
  default: return "not stored";
  }
}

/* Convers a quality_t enum to a cstr. */
char *item_qual_to_cstr(const enum quality_t qual) {
  switch(qual) {
  case normal_q: return "normal"; break;
  case low_q: return "low"; break;
  case high_q: return "high"; break;
  case magic_q: return "magic"; break;
  case set_q: return "set"; break;
  case unique_q: return "unique"; break;
  case rare_q: return "rare"; break;
  case crafted_q: return "crafted"; break;
  default: return "unknown"; break;
  }
}

/* The rest of this code isn't important or used, and isn't commented. */
/* Prints affix IDs. */
void print_affix_set(const uint16 *affix, const unsigned int pos) {
  if(affix[pos]) {
    printf("(prefix 0x%04x", affix[pos]);
    if(affix[pos + 1]) {
      printf(", "); }
  }
  if(affix[pos + 1]) {
    if(!affix[pos]) {
      printf("("); }
    printf("suffix 0x%04x", affix[pos + 1]);
  }
  printf(")");
}

/* Print item information for non-simple items. */
void print_extended_item(struct item_t *item, const unsigned int clvl, const uint8 *set_info) {
  unsigned int pos, set_id, bits_set;
  switch(item->qual) {
  case low_q:
    printf("low quality (qual %d) ", item->subqual);
    break;
  case normal_q:
    break;
  case high_q:
    printf("high quality (qual %d) ", item->subqual);
    break;
  case magic_q:
    printf("magic (blue) ");
    print_affix_set(item->affix, 0);
    printf(" ");
    break;
  case set_q:
    printf("set (green) (set 0x%04x) (number %d) ", item->suid, lookup_set_item_id_by_suid(item->suid));
    break;
  case unique_q:
    printf("unique (gold) (uid 0x%04x) ", item->suid);
    break;
  case rare_q:
    printf("rare (yellow) (names 0x%02x 0x%02x) ", item->name1, item->name2);
    for(pos = 0; pos < 6; pos += 2) {
      print_affix_set(item->affix, pos);
      printf(" "); }
    break;
  case crafted_q:
    printf("crafted (orange) (names 0x%02x 0x%02x) ", item->name1, item->name2);
    for(pos = 0; pos < 6; pos += 2) {
      print_affix_set(item->affix, pos);
      printf(" "); }
    break;
  default:
    printf("unknown ");
    break;
  }
  if(item->has_runeword == yes) {
    if(lookup_rune_by_id(item->runeword) && lookup_string_by_name(lookup_rune_by_id(item->runeword))) {
      printf("with runeword \"%s\" ", lookup_string_by_name(lookup_rune_by_id(item->runeword))); }
    else {
      printf("with unknown runeword (id 0x%04x) ", item->runeword); }
  }
  if(item->has_charm_affix == yes) {
    printf(item->charm_suff == yes ? "(charm suffix 0x%04x) " : "(charm prefix 0x%04x) ", item->charm_affix); }
  if(item->has_class_info == yes) {
    printf("(class-specific tag 0x%04x) ", item->class_info); }
  if(item->has_alt_gfx == yes) {
    printf("graphic id 0x%02x ", item->alt_gfx); }
  printf("ilvl %d GUID 0x%08x ", item->ilvl, item->guid);
  if(item->max_dura) {
    printf("dura %d / %d ", item->dura, item->max_dura); }
  if(item->def) {
    printf("%d defense ", item->def); }
  printf("\n");
  if(print_prop_list(item->norm_pl, clvl, "\n")) {
    printf("\n"); }
  if(item->qual == set_q) {
    set_id = lookup_set_id_by_suid(item->suid);
    bits_set = count_bits(set_info[set_id]);
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos]) {
	if(bits_set <= pos + 1) {
	  printf("x "); }
	printf("(%d) ", pos + 1);
	if(print_prop_list(item->set_pl[pos], clvl, ", ")) {
	  printf("\n"); }
      }
    }
  }
  for(pos = 0; pos < item->sockets_used; pos++) {
    printf("in socket %d: ", pos);
    print_item_name(item->socketed_items[pos]);
    printf("\n");
  }
}

/* Prints an item's identifier, quality, etc. */
void print_item_name(struct item_t *item) {
  char *name, *name1, *name2;
  if(item->is_ear == yes) {
    printf("Ear (%s, lvl %d %s)", (item->ear_info).name, (item->ear_info).level, char_class_to_cstr((item->ear_info).ear_class));
  }
  else if(item->name) {
    name = lookup_string_by_name(item->tag) ? lookup_string_by_name(item->tag) : item->name;
    switch(item->qual) {
    case low_q:
      printf("Low Quality %s (0x%02x)", name, item->subqual);
      break;
    case high_q:
      printf("High Quality %s (0x%02x)", name, item->subqual);
      break;
    case normal_q:
      printf("%s", name);
      break;
    case magic_q:
      if(item->affix[0] && lookup_magic_pre_by_id(item->affix[0])) {
	printf("%s ", lookup_magic_pre_by_id(item->affix[0])); }
      printf("%s", name);
      if(item->affix[1] && lookup_magic_suf_by_id(item->affix[1])) {
	printf(" %s", lookup_magic_suf_by_id(item->affix[1])); }
      printf(" (blue)");
      break;
    case rare_q:
    case crafted_q:
      if(lookup_rare_pre_by_id(item->name1) && lookup_rare_suf_by_id(item->name2)) {
	name1 = lookup_string_by_name(lookup_rare_pre_by_id(item->name1)) ? lookup_string_by_name(lookup_rare_pre_by_id(item->name1)) : lookup_rare_pre_by_id(item->name1);
	name2 = lookup_string_by_name(lookup_rare_suf_by_id(item->name2)) ? lookup_string_by_name(lookup_rare_suf_by_id(item->name2)) : lookup_rare_suf_by_id(item->name2);
	printf("%s %s (%s) ", name1, name2, name); }
      else {
	printf("%s ", name); }
      printf(item->qual == rare_q ? "(yellow)" : "(orange)");
      break;
    case set_q:
      if(lookup_set_item_by_id(item->suid)) {
	name1 = lookup_string_by_name(lookup_set_item_by_id(item->suid)->name) ? lookup_string_by_name(lookup_set_item_by_id(item->suid)->name) : lookup_set_item_by_id(item->suid)->name;
	printf("%s (%s) (green)", name1, name); }
      else {
	printf("%s (green)", name); }
      break;
    case unique_q:
      if(item->suid == 0xfff) {
	printf("Standard of Heroes (small charm) (gold)"); }
      else if(lookup_unique_by_id(item->suid)) {
	name1 = lookup_string_by_name(lookup_unique_by_id(item->suid)->name) ? lookup_string_by_name(lookup_unique_by_id(item->suid)->name) : lookup_unique_by_id(item->suid)->name;
	printf("%s (%s) (gold)", name1, name); }
      else {
	printf("%s (gold)", name); }
      break;
    default:
      printf("%s", name);
      break;
    }
  }
  else {
    printf("unknown tag \"%s\"", item->tag); }
}

/* Prints an item's location. */
void print_item_loc(struct item_t *item) {
  if(item->location == equipped) {
    printf(item->worn == right_hand || item->worn == left_hand || item->worn == alt_right_hand || item->worn == alt_left_hand ? "used in %s" : "worn on %s", item_worn_to_cstr(item->worn)); }
  else if(item->location == stored) {
    printf("stored at (%d, %d) in %s", item->row, item->col, item_grid_loc_to_cstr(item->grid_loc)); }
  else if(item->location == socketed) {
    printf("socketed in ");
    if(item->contained_in) {
      print_item_name(item->contained_in); }
    else {
      printf("(unknown)"); }
  }
  else {
    printf(item_location_to_cstr(item->location)); }
  if(item->section != player) {
    printf(" (%s)", item_section_to_cstr(item->section)); }
}

/* Prints various item flags.  */
unsigned int print_item_flags(struct item_t *item) {
  unsigned int flags_printed = 0;
  if(item->is_identified == no) {
    printf(flags_printed++ ? ", not id'd" : "not id'd"); }
  if(item->is_disabled == yes) {
    printf(flags_printed++ ? ", disabled" : "disabled"); }
  if(item->is_socketed == yes) {
    if(item->sockets != 0) {
      printf(flags_printed++ ? ", socketed (%d / %d)" : "socketed (%d / %d)", item->sockets_used, item->sockets); }
    else {
      printf(flags_printed++ ? ", socketed (%d / ?)" : "socketed (%d / ?)", item->sockets_used); }
  }
  if(item->is_illegal == yes) {
    printf(flags_printed++ ? ", illegal item" : "illegal item"); }
  if(item->is_ear == yes) {
    printf(flags_printed++ ? ", ear" : "ear"); }
  if(item->is_newbie == yes) {
    printf(flags_printed++ ? ", newbie" : "newbie"); }
  /*if(item->simple == yes) {
  printf(flags_printed++ ? ", simple item" : "simple item"); } */
  if(item->is_ethereal == yes) {
    printf(flags_printed++ ? ", ethereal" : "ethereal"); }
  if(item->is_personalized == yes) {
    if(item->inscription[0] != '\0') {
      printf(flags_printed++ ? ", personalized as \"%s\"" : "personalized as \"%s\"", item->inscription); }
    else {
      printf(flags_printed++ ? ", personalized" : "personalized"); }
  }
  if(item->has_runeword == yes) {
    printf(flags_printed++ ? ", has runeword" : "has runeword"); }
  if(flags_printed) {
    printf("\n"); }
  return flags_printed;
}

/* Prints an item, including its raw bytes. */
void print_item(struct item_t *item, const unsigned int clvl, const uint8 *set_info) {
  unsigned int byte_pos;
  print_item_name(item);
  printf(" - ");
  print_item_loc(item);
  printf("\n");
  if(item->location == socketed) {
    printf("|   "); }
  if(print_item_flags(item) && item->location == socketed) {
    printf("|   "); }
  if(item->is_simple == no) {
    print_extended_item(item, clvl, set_info); }
  for(byte_pos = 0; byte_pos < item->size; byte_pos++) {
    printf(byte_pos == item->size - 1 ? "%02X\n" : "%02X ", item->bytes[byte_pos]);
  }
}

/* Print global set bonuses. */
void print_set_bonuses(const struct charinfo_t *charinfo) {
  unsigned int set_no, set_count, pos;
  int bonus = -1;
  char *name;
  enum flag_t bonus_printed = no;
  struct set_entry_t *set;
  for(set_no = 0; set_no < sizeof_set; set_no++, bonus = -1) {
    if(charinfo->set_info[set_no]) {
      set = lookup_set_by_sid(set_no);
      set_count = count_bits(charinfo->set_info[set_no]);
      if(set_count == set->items) {
	bonus = 4; }
      else {
	for(pos = 0; pos < 4; pos++) {
	  if(set->set_bonus[pos] && set_count > pos + 1) {
	    bonus = pos; }
	}
      }
      if(bonus != -1) {
	if(bonus_printed == no) {
	  bonus_printed = yes;
	  printf("Set bonuses:\n"); }
	name = set->name;
	if(lookup_string_by_name(name)) {
	  name = lookup_string_by_name(name); }
	printf("%s (%d/%d)\n", name, set_count, set->items);
	if(print_prop_list(set->set_bonus[bonus], (charinfo->stats).level, "\n")) {
	  printf("\n"); }
      }
    }
  }
}
