#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define __USE_XOPEN
#include <time.h>
#include <sys/types.h>

#ifdef WIN32
#include "dirent.h"
#else
#include <dirent.h>
#endif

#include "d2stats.h"

/* Extracts a seconds-since-epoch value from a cstring of the form
   'Fri Aug 22 20:30:21 MDT'. Does nothing on Windows. */
#ifdef WIN32
time_t get_time_from_cstr(const char *time_buf) {
  return 0; }
#else
time_t get_time_from_cstr(const char *time_buf) {
  time_t curr_time, dummy;
  struct tm *time_struct;
  curr_time = time(&dummy);
  time_struct = localtime(&curr_time);
  strptime(time_buf, "%a %b %d %H:%M:%S %Z", time_struct);
  return mktime(time_struct);
}
#endif

/* Reads a PvPGN flat report file line-by-line to extract information
   that goes into 'game'. */
unsigned int extract_report(FILE *file, struct game_t *game) {
  char buf[1024], time_buf[1024];
  unsigned int tokens = 0, len, joins = 0;
  while(fgets(buf, 1024, file) && tokens < 5) {
    if(!strncmp(buf, "name=\"", 6)) {
      for(len = 0; len < strlen(buf + 6) && buf[len + 6] != '\"'; len++);
      if(len != 0) {
	strncpy(game->name, buf + 6, strlen(buf + 6) > 63 ? 63 : strlen(buf + 6));
	game->name[len] = '\0';
	tokens++;
      }
    }
    else if(!strncmp(buf, "created=\"", 9) && strlen(buf) > 76 && !strncmp(buf + 68, "ended=\"", 7)) {
      /* copy both the start and end times from this line. */
      for(len = 0; len < strlen(buf + 9) && buf[len + 9] != '\"'; len++);
      if(len != 0) {
	strncpy(time_buf, buf + 9, len);
	time_buf[len] = '\0';
	game->start_time = get_time_from_cstr(time_buf);
	tokens++;
      }
      for(len = 0; len < strlen(buf + 75) && buf[len + 75] != '\"'; len++);
      if(len != 0) {
	strncpy(time_buf, buf + 75, len);
	time_buf[len] = '\0';
	game->end_time = get_time_from_cstr(time_buf);
	tokens++;
      }
    }
    else if(!strncmp(buf, "joins=", 6)) {
      sscanf(buf + 6, "%d", &joins);
      game->players = joins;
      tokens++;
    }
    else if(!strncmp(buf, "clienttag=", 10)) {
      if(strncmp(buf + 10, "D2XP", 4)) {
	game->not_d2 = yes;
	return tokens; }
      game->not_d2 = no;
      tokens++;
    }
  }
  return tokens;
}

/* Attempts to move a PvPGN report at 'dir'/'file' to its backup directory
   as specified in 'report_backup' in the config. Returns yes if successful,
   no otherwise. */
enum flag_t move_report(const char *file, const char *dir) {
  static char buf[1024] = "", src[1024], dst[1024];
  if(buf[0] == '\0') {
    config(report_backup_path, buf);
    if(buf[0] == '\0') {
      return no; }
  }
  strncpy(src, dir, strlen(dir) + 1);
  strncat(src, DIR_SLASH, 2);
  strncat(src, file, strlen(file) + 1);
  strncpy(dst, buf, strlen(buf) + 1);
  strncat(dst, DIR_SLASH, 2);
  strncat(dst, file, strlen(file) + 1);
  if(rename(src, dst)) {
    write_log("couldn't move \"%s\" to \"%s\"\n", 3, src, dst);
    return no; }
  return yes;
}

/* Reads every report file in the directory in 'report' in config
   to the double-arr pointed to by 'games'. Returns the number of successful
   extractions. */
unsigned int read_reports(struct game_t ***games) {
  unsigned int rept_no, count;
  DIR *dir;
  FILE *file;
  struct dirent *dp;
  char buf[1024] = "", filename[1024];
  enum flag_t log_this = no, move_files = yes;
  config(report_path, buf);
  if(buf[0] == '\0' || !(dir = opendir(buf))) {
    return 0; }
  for(count = 0, dp = readdir(dir); dp; dp = readdir(dir)) {
    if(!strncmp(dp->d_name, "gr_", 3)) {
      count++; }
  }
  closedir(dir);
  /* Reports pile up fast. Let the user know what's going on. */
  if(count > 5000) {
    log_this = yes;
    write_log("%d reports found, reading may take a while... ", 3, count); }
  else if(count == 0) {
    return 0; }
  if(!(*games = malloc(count * sizeof(struct game_t *)))) {
    return 0; }
  zero_out(*games, count * sizeof(struct game_t *)); 
  dir = opendir(buf);
  for(rept_no = 0, dp = readdir(dir); dp && rept_no < count; dp = readdir(dir)) {
    if(!strncmp(dp->d_name, "gr_", 3)) {
      /* Every 1000 reports, print out debug output. */
      if(log_this == yes && rept_no && rept_no % 1000 == 0) {
	write_log("%d... ", 3, rept_no); }
      strncpy(filename, buf, strlen(buf) + 1);
      strncat(filename, DIR_SLASH, 2);
      strncat(filename, dp->d_name, strlen(dp->d_name) + 1);
      if((file = fopen(filename, "r"))) {
	if(!((*games)[rept_no] = malloc(sizeof(struct game_t)))) {
	  return 0; }
	zero_out((*games)[rept_no], sizeof(struct game_t));
	extract_report(file, (*games)[rept_no]);
	fclose(file);
	/* If every move has been successful so far, try another. */
	if(move_files == yes) {
	  move_files = move_report(dp->d_name, buf); }
      }
      rept_no++;
    }
  }
  closedir(dir);
  if(log_this == yes) {
    write_log("done.\n", 3); }
  return count;
}

/* Adds 'game' to the tally at 'list', using 'curr_time' to determine whether
   it counts as a recent game. */
void tally_game(struct list_t *list, const struct game_t *game, const unsigned int curr_time) {
  /* 5 days = recent. 30 mintes = game over the limit, 'significant'. */
  struct tm *tmp;
  const unsigned int recent = 3600 * 24 * 5, lim = 1800;
  const unsigned int length = (unsigned int)(game->end_time - game->start_time);
  unsigned int pos;
  /* 'mod' used to be a weight for averaging a game's significance. as of now
     the number of players is simply used, but you could do stuff like,
     multiply by length to weight a game's 'significance' other ways. */
  double mod = game->players;
  tmp = localtime(&game->start_time);
  list->games++;
  if(length >= lim) {
    /* Only total players in games that are 'significant', or the average is
       going to be very close to 1 thanks to solo MF runs. */
    list->avg_players += game->players;
    list->games_over_lim++; }
  /* For every hour the game has lasted, add 'mod' to that hour. */
  for(pos = 0; pos * 3600 < length; pos++) {
    list->avg_game_time[(tmp->tm_hour + pos) % 24] += mod; }
  list->avg_game_time_mod += mod;
  list->avg_game_len += length;
  list->total_game_len += length;
  /* Same as above but for games that are 'recent'. */
  if(curr_time - game->start_time <= recent) {
    list->recent++;
    if(length >= lim) {
      list->recent_avg_players += game->players;
      list->recent_over_lim++; }
    for(pos = 0; pos * 3600 < length; pos++) {
      list->recent_avg_time[(tmp->tm_hour + pos) % 24] += mod; }
    list->recent_avg_time_mod += mod;
    list->recent_avg_len += length;
    list->recent_total_len += length;
  }
}

/* Appends data contained in 'games' to 'report_log' in the config using a
   simple format: '"name" start_epoch end_epoch length_in_sec players\n' */
void write_games(struct game_t **games, const unsigned int count) {
  FILE *file;
  char buf[1024] = "";
  unsigned int game_no;
  config(report_log, buf);
  if(buf[0] != '\0' && count) {
    if(!(file = fopen(buf, "a"))) {
      write_log("could not open file %s\n", 3, buf);
      return; }
    for(game_no = 0; game_no < count; game_no++) {
      if(games[game_no]->not_d2 == no) {
	fprintf(file, "\"%s\" %d %d %d %d\n", games[game_no]->name, (int)games[game_no]->start_time, (int)games[game_no]->end_time, (int)(games[game_no]->end_time - games[game_no]->start_time), games[game_no]->players);
      }
    }
    fclose(file);
  }
}

/* Reads games in from the file specified by 'report_log' in the config, if
   it exists and if 'read_report_log' is 'yes'. */
unsigned int read_games(struct game_t ***games, unsigned int count) {
  FILE *file;
  char buf[1024] = "", buf2[1024] = "", line[1024];
  unsigned int len, game_no, dummy, toklen;
  struct game_t *game;
  config(report_log, buf);
  config(read_report_log, buf2);
  if(buf[0] != '\0' && !strncmp(buf2, "yes", 4)) {
    if(!(file = fopen(buf, "r"))) {
      return count; }
    /* Line count. */
    for(len = 0; fgets(line, 1024, file);) {
      if(line[0] == '\"') {
	len++; }
    }
    rewind(file);
    if(!len) {
      fclose(file);
      return count; }
    /* If games has already been alloc'd, realloc. */
    if(count) {
      if(!(*games = realloc(*games, (count + len) * sizeof(struct game_t *)))) {
	fclose(file);
	return count; }
      zero_out(*games + count, len * sizeof(struct game_t *));
    }
    else {
      if(!(*games = malloc(len * sizeof(struct game_t *)))) {
	fclose(file);
	return count; }
      zero_out(*games, len * sizeof(struct game_t *));
    }
    for(game_no = 0; fgets(line, 1024, file) && game_no < len;) {
      if(!(game = malloc(sizeof(struct game_t)))) {
	return count; }
      if(line[0] == '\"') {
	/* Get length of the name within double-quotes and copy it. */
	for(toklen = 0; line[toklen + 1] != '\0' && line[toklen + 1] != '\"'; toklen++);
	strncpy(game->name, line + 1, toklen);
	game->name[toklen] = '\0';
	if(strlen(line) > toklen + 3) {
	  sscanf(line + toklen + 3, "%d %d %d %d", (int *)&game->start_time, (int *)&game->end_time, &dummy, &game->players);
	  game->not_d2 = no; }
	else {
	  /* This shouldn't happen, but just in case... */
	  game->not_d2 = yes; }
	(*games)[game_no + count] = game;
	game_no++;
      }
    }
    count += len;
  }
  return count;
}

/* Reads all games from all sources and tallies them into 'list'. */
void parse_games(struct list_t *list) {
  unsigned int count, game_no, curr_time, pos, extra;
  char buf[1024] = "";
  time_t dummy;
  struct game_t **games;
  struct tm *tmp;
  config(parse_reports, buf);
  curr_time = (unsigned int)time(&dummy);
  if(!strncmp(buf, "yes", 4)) {
    count = read_reports(&games);
    extra = read_games(&games, count);
    if(!count && !extra) {
      write_log("could not open report files\n", 3);
      return; }
    if(count) {
      write_games(games, count); }
    count = extra;
    for(game_no = 0; game_no < count; game_no++) {
      if(games[game_no] && games[game_no]->not_d2 == no) {
	tally_game(list, games[game_no], curr_time); }
    }
    if(list->games && list->avg_game_len && list->avg_players) {
      /* Compute averages. */
      for(pos = 0; pos < 24; pos++) {
	list->avg_game_time[pos] /= list->avg_game_time_mod; }
      list->avg_game_len /= list->games;
      list->avg_players /= list->games_over_lim;
    }
    if(list->recent && list->recent_avg_len && list->recent_avg_players) {
      for(pos = 0; pos < 24; pos++) {
	list->recent_avg_time[pos] /= list->recent_avg_time_mod; }
      list->recent_avg_len /= list->recent;
      list->recent_avg_players /= list->recent_over_lim;
    }
    for(game_no = 0; game_no < count; game_no++) {
      free(games[game_no]);
    }
    free(games);
    /* Get the timezone, for display. */
    tmp = localtime((time_t *)&curr_time);
    if(tmp->tm_isdst) {
      strncpy(list->timezone, tzname[1], strlen(tzname[1]) + 1); }
    else {
      strncpy(list->timezone, tzname[0], strlen(tzname[0]) + 1); }
  }
}
