#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"
#include "tally.h"
#include "hashtable.h"
#include "hashtable_itr.h"

/* Returns the extra mana provided by energy, which depends on class. */
unsigned int lookup_mana_from_ener(const unsigned int ener, const enum char_class_t char_class) {
  double val;
  switch(char_class) {
  case ama: val = ener * 1.5; break;
  case asn: val = ener * 1.75; break;
  case nec: val = ener * 2.0; break;
  case bar: val = ener * 1.0; break;
  case src: val = ener * 2.0; break;
  case dru: val = ener * 2.0; break;
  case pal: val = ener * 1.5; break;
  default: val = 0; break;
  }
  return (unsigned int)floor(val + 0.5);
}

/* Returns the extra life provided by vitality, which depends on class. */
unsigned int lookup_life_from_vit(const unsigned int vit, const enum char_class_t char_class) {
  double val;
  switch(char_class) {
  case ama: val = vit * 3.0; break;
  case asn: val = vit * 3.0; break;
  case nec: val = vit * 2.0; break;
  case bar: val = vit * 4.0; break;
  case src: val = vit * 2.0; break;
  case dru: val = vit * 2.0; break;
  case pal: val = vit * 3.0; break;
  default: val = 0; break;
  }
  return (unsigned int)floor(val + 0.5);
}

/* Returns the extra stam provided by vitality, which depends on class. */
unsigned int lookup_stam_from_vit(const unsigned int vit, const enum char_class_t char_class) {
  double val;
  switch(char_class) {
  case ama: val = vit * 1.0; break;
  case asn: val = vit * 1.25; break;
  case nec: val = vit * 1.0; break;
  case bar: val = vit * 1.0; break;
  case src: val = vit * 1.0; break;
  case dru: val = vit * 1.0; break;
  case pal: val = vit * 1.0; break;
  default: val = 0; break;
  }
  return (unsigned int)floor(val + 0.5);
}

/* Used by tally_props to check to see whether or not a class ID (in prop 83
   '+3 to Sorceress Class Skills' or prop 188 '+3 to Fire Skills (Sorceress
   Only') matches a certain char_class_t enum. */
enum flag_t compare_class(const unsigned int c, const enum char_class_t char_class) {
  switch(char_class) {
  case ama: return c == 0 ? yes : no;
  case src: return c == 1 ? yes : no;
  case nec: return c == 2 ? yes : no;
  case pal: return c == 3 ? yes : no;
  case bar: return c == 4 ? yes : no;
  case dru: return c == 5 ? yes : no;
  case asn: return c == 6 ? yes : no;
  default: return no;
  }
}

/* Returns the offset into the 'skilldesc.txt' table, where the start entry
   for the 30 skills for 'char_class' can be found. */
unsigned int class_skill_offs(const enum char_class_t char_class) {
  switch(char_class) {
  case ama: return 6;
  case src: return 36;
  case nec: return 66;
  case pal: return 96;
  case bar: return 126;
  case dru: return 221;
  case asn: return 251;
  default: return 0; }
}

/* Returns a character's resistance modifier. A character has +10 resistance
   to everything for every Scroll of Resistance they've read. A character has
   -40 resistance all in nightmare difficulty, and -100 resistance all in hell
   difficulty. As an aside, resistance is capped at 75 with no +max,
   and +max is capped at 95, but this doesn't apply to PvP combat. */
int get_res_mod(const struct charinfo_t *charinfo) {
  int res_mod = 0, pos;
  for(pos = 0; pos < 3; pos++) {
    if(charinfo->has_res[pos] == yes) {
      res_mod += 10; }
  }
  switch(charinfo->char_title) {
  case sir: res_mod -= 40; break;
  case lord:
  case baron: res_mod -= 100; break;
  default: break;
  }
  return res_mod;
}

/* Frees the totals proplist.*/
void free_total(struct prop_t *total) {
  struct prop_t *clean, *tmp;
  for(clean = total; clean != NULL;) {
    tmp = clean->next;
    free(clean);
    clean = tmp;
  }
}

/* Prints the character's skill spec to 'buf'. One simple example is
   '5/40/10', meaning, 5 skill points in the first tab, 40 in the second,
   10 in the third. */
void get_spec(const struct charinfo_t *charinfo, char *buf) {
  unsigned int tab1 = 0, tab2 = 0, tab3 = 0, pos = 0;
  const uint8 *skills = (charinfo->stats).skills;
  for(pos = 0; pos < 30; pos++) {
    switch(charinfo->char_class) {
    /* These numbers won't make a lot of sense by themselves. They are offsets
       into the character's skill array, and they match entries in the
       skilldesc.txt game table. There are a lot of them (210) so I won't list
       them here; open that table if you need to know what numbers do what.
       It isn't immediately obvious which skills belong to which skill tab
       unless you've played the game yourself for more than two minutes. */
    case ama:
      switch(pos) {
      case  0: case  1: case  5: case  6: case 10:
      case 15: case 16: case 20: case 21: case 25: tab1 += skills[pos]; break;
      case  2: case  3: case  7: case 11: case 12:
      case 17: case 22: case 23: case 26: case 27: tab2 += skills[pos]; break;
      case  4: case  8: case  9: case 13: case 14:
      case 18: case 19: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case src:
      switch(pos) {
      case  0: case  1: case  5: case 10: case 11:
      case 15: case 16: case 20: case 25: case 26: tab1 += skills[pos]; break;
      case  2: case  6: case  7: case 12: case 13:
      case 17: case 18: case 21: case 22: case 27: tab2 += skills[pos]; break;
      case  3: case  4: case  8: case  9: case 14:
      case 19: case 23: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case nec:
      switch(pos) {
      case  0: case  5: case  6: case 10: case 11:
      case 15: case 16: case 20: case 21: case 25: tab1 += skills[pos]; break;
      case  1: case  2: case  7: case  8: case 12:
      case 17: case 18: case 22: case 26: case 27: tab2 += skills[pos]; break;
      case  3: case  4: case  9: case 13: case 14:
      case 19: case 23: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case pal:
      switch(pos) {
      case  0: case  1: case  5: case 10: case 11:
      case 15: case 16: case 20: case 21: case 25: tab1 += skills[pos]; break;
      case  2: case  6: case  7: case 12: case 17:
      case 18: case 22: case 23: case 26: case 27: tab2 += skills[pos]; break;
      case  3: case  4: case  8: case  9: case 13:
      case 14: case 19: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case bar:
      switch(pos) {
      case  0: case  6: case  7: case 13: case 14:
      case 17: case 18: case 21: case 25: case 26: tab1 += skills[pos]; break;
      case  1: case  2: case  3: case  8: case  9:
      case 10: case 15: case 19: case 22: case 27: tab2 += skills[pos]; break;
      case  4: case  5: case 11: case 12: case 16:
      case 20: case 23: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case dru:
      switch(pos) {
      case  0: case  1: case  5: case  6: case 10:
      case 15: case 16: case 20: case 25: case 26: tab1 += skills[pos]; break;
      case  2: case  3: case  7: case 11: case 12:
      case 17: case 18: case 21: case 22: case 27: tab2 += skills[pos]; break;
      case  4: case  8: case  9: case 13: case 14:
      case 19: case 23: case 24: case 28: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    case asn:
      switch(pos) {
      case  0: case  5: case  6: case 10: case 11:
      case 15: case 20: case 21: case 25: case 26: tab1 += skills[pos]; break;
      case  1: case  2: case  7: case 12: case 13:
      case 16: case 17: case 22: case 27: case 28: tab2 += skills[pos]; break;
      case  3: case  4: case  8: case  9: case 14:
      case 18: case 19: case 23: case 24: case 29: tab3 += skills[pos]; break;
      default: break;
      }
      break;
    default:
      break;
    }
  }
  sprintf(buf, "%d/%d/%d", tab1, tab2, tab3);
}

/* Adds an ear to the 'list->ear_list' hashtable, creating it if it doesn't
   already exist. The table merely keeps a tally of what ear names have how
   many ears floating around. */
void add_ear(struct list_t *list, struct ear_t *ear) {
  unsigned int *ears;
  char *name;
  if(!list->ear_list) {
    list->ear_list = create_hashtable(50, hash_cstr, hash_cmp_cstr);
  }
  /* if the ear name is already in the table, just increment it. */
  if(hashtable_search(list->ear_list, ear->name)) {
    ears = (unsigned int *)hashtable_search(list->ear_list, ear->name);
    (*ears)++; }
  /* otherwise, malloc both a key (ear name) and value (1) and insert. */
  else {
    if(!(ears = malloc(sizeof(unsigned int)))) {
      return; }
    *ears = 1;
    if(!(name = malloc(strlen(ear->name) + 1))) {
      return; }
    strncpy(name, ear->name, strlen(ear->name));
    name[strlen(ear->name)] = '\0';
    hashtable_insert(list->ear_list, name, ears);
  }
}

/* Adds a prop's damage enhancement to the weapon bonus damage tally. If the
   prop made an impact on the damage modifiers, returns yes. */
enum flag_t add_to_dmg(struct prop_t *prop, double *mult_min, double *mult_max, unsigned int *add_min, unsigned int *add_max) {
  if(prop->id == 17) { /* damage multiplier - always appears in min/max pair */
    *mult_min += prop->min / 100.0;
    *mult_max += prop->max / 100.0;
    return yes; }
  else if(prop->id == 21) { /* +min */
    *add_min += prop->val;
    return yes; }
  else if(prop->id == 22) { /* +max */
    *add_max += prop->val;
    return yes; }
  return no;
}

/* Adds a prop's defense enhancement to the armor bonus tally. If the prop
   made an impact on the armor modifiers, returns yes. */
enum flag_t add_to_def(struct prop_t *prop, double *mult, unsigned int *add) {
  if(prop->id == 16) { /* defense multiplier */
    *mult += prop->val / 100.0;
    return yes; }
  else if(prop->id == 31) {
    *add += prop->val; /* +defense */
    return yes; }
  return no;
}

/* Adds a prop's block enhancement to the block bonus tally. If the prop made
   an impact on the block modifer, returns yes. */
enum flag_t add_to_block(struct prop_t *prop, unsigned int *add) {
  if(prop->id == 20) { /* +block */
    *add += prop->val;
    return yes; }
  return no;
}

/* Calculates an item's bonus damage (min / max add and multiply bonuses) and
   returns yes if a bonus exists. Considers active set bonuses too. */
enum flag_t get_dmg(struct item_t *item, double *mult_min, double *mult_max, unsigned int *add_min, unsigned int *add_max, const uint8 *set_info) {
  unsigned int pos, set_no, bits_set;
  struct prop_t *prop;
  enum flag_t is_enhanced = no;
  *mult_min = 1.0;
  *mult_max = 1.0;
  *add_min = 0;
  *add_max = 0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    if(add_to_dmg(prop, mult_min, mult_max, add_min, add_max) == yes) {
      is_enhanced = yes; }
  }
  if(item->qual == set_q) {
    set_no = lookup_set_id_by_suid(item->suid);
    bits_set = count_bits(set_info[set_no]);
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos] && bits_set > pos + 1) {
	for(prop = item->set_pl[pos]; prop != NULL; prop = prop->next) {
	  if(add_to_dmg(prop, mult_min, mult_max, add_min, add_max) == yes) {
	    is_enhanced = yes; }
	}
      }
    }
  }
  return is_enhanced;
}
    
/* Calculates an item's bonus armor (add and multiply bonuses) and returns yes
   if it exists. Considers active set bonuses too. */
enum flag_t get_def(struct item_t *item, double *mult, unsigned int *add, const uint8 *set_info) {
  unsigned int pos, set_no, bits_set;
  struct prop_t *prop;
  enum flag_t is_enhanced = no;
  *mult = 1.0;
  *add = 0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    if(add_to_def(prop, mult, add) == yes) {
      is_enhanced = yes; }
  }
  if(item->qual == set_q) {
    set_no = lookup_set_id_by_suid(item->suid);
    bits_set = count_bits(set_info[set_no]);
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos] && bits_set > pos + 1) {
	for(prop = item->set_pl[pos]; prop != NULL; prop = prop->next) {
	  if(add_to_def(prop, mult, add) == yes) {
	    is_enhanced = yes; }
	}
      }
    }
  }
  return is_enhanced;
}
    
/* Calculates an item's bonus block (additive bonuses) and returns yes if it
   exists. Considers active set bonuses too. */
enum flag_t get_block(struct item_t *item, unsigned int *add, const uint8 *set_info) {
  unsigned int pos, set_no, bits_set;
  struct prop_t *prop;
  enum flag_t is_enhanced = no;
  *add = 0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    if(add_to_block(prop, add) == yes) {
      is_enhanced = yes; }
  }
  if(item->qual == set_q) {
    set_no = lookup_set_id_by_suid(item->suid);
    bits_set = count_bits(set_info[set_no]);
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos] && bits_set > pos + 1) {
	for(prop = item->set_pl[pos]; prop != NULL; prop = prop->next) {
	  if(add_to_block(prop, add) == yes) {
	    is_enhanced = yes; }
	}
      }
    }
  }
  return is_enhanced;
}

/* Calculates an item's bonus lowered requirement modifier and returns yes
   if it exists. Does not consider set bonuses, because no set has them
   (if you played the game, think about why that is for a second). */
enum flag_t get_req(struct item_t *item, double *mult) {
  struct prop_t *prop;
  enum flag_t is_enhanced = no;
  *mult = 1.0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    if(prop->id == 91) { /* 'Requirements -20%' */
      *mult += prop->val / 100.0; 
      is_enhanced = yes; }
  }
  return is_enhanced;
}

/* Calculates an item's bonus durability modifiers (additive and
   multiplicative) and returns yes if they exist. Does not consider set
   bonuses; again, no set has them. Wouldn't make a lot of sense for a set
   item to suddenly gain durability (meaning it's no longer broken) if you
   put on extra pieces of the set... */
enum flag_t get_dura(struct item_t *item, double *mult, unsigned int *add) {
  struct prop_t *prop;
  enum flag_t is_enhanced = no;
  *mult = 1.0;
  *add = 0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    if(prop->id == 73) { /* +dura */
      *add += prop->val;
      is_enhanced = yes; }
    if(prop->id == 75) { /* 'Increases Maximum Durability 50%' */
      *mult += prop->val / 100.0; 
      is_enhanced = yes; }
    else if(prop->id == 152) { /* 'Indestructible' - max_dura == 0 */
      *mult = 0.0;
      is_enhanced = yes; }
  }
  return is_enhanced;
}

/* Arbitrary-as-hell method to calculate an estimated 'value' for a property
   that's attached to an item. Uses 'ItemStatCost.txt' columns 'multiply', 
   'add' and 'divide' columns in that order, sort of. */
double calc_prop_value(struct prop_t *prop) {
  struct property_entry_t *entry, *entry2, *entry3;
  double val = 0.0;
  if((entry = lookup_prop_by_id(prop->id)) && entry->add) {
    /* min/max/length pairs, as always, are special cases */
    if(prop->min && prop->max) {
      switch(prop->id) {
      case 17:
      case 21:
      case 23:
      case 48:
      case 50:
      case 52:
	/* if they occur together, add both min and max values by their
	   respective table entries, and cap it at 20 */
	if((entry2 = lookup_prop_by_id(prop->id + 1)) && entry2->add) {
	  val += (prop->min * entry->mult + entry->add) / (double)entry->div;
	  val += (prop->max * entry2->mult + entry2->add) / (double)entry2->div;
	  return val > 20 ? 20 : val;
	}
	break;
      case 54:
      case 57:
	/* if cold/pois min/max/len occurs together (there's no reason why
	   they shouldn't) add the total value of all three props, and cap it
	   at 20 */
	if((entry2 = lookup_prop_by_id(prop->id + 1)) && entry2->add) {
	  if((entry3 = lookup_prop_by_id(prop->id + 2)) && entry3->add) {
	  val += (prop->min * entry->mult + entry->add) / (double)entry->div;
	  val += (prop->max * entry2->mult + entry2->add) / (double)entry2->div;
	  val += (prop->length * entry3->mult + entry3->add) / (double)entry3->div;
	  return val > 20 ? 20 : val;
	  }
	}
	break;
      default: break;
      }
    }
    /* charged / otherwise skills tend to be extremely valuable in the table,
       so use the skill level to do the computation and cap it at 15 */
    else if(prop->skill_level) {
      val += (prop->skill_level * entry->mult + entry->add) / (double)entry->div;
      return val > 15 ?  15 : val; }
    val += (prop->val * entry->mult + entry->add) / (double)entry->div;
    /* the following props are EXTREMELY unbalanced. class skills (83) has an
       add column of nearly 50,000 and a multiply column of 1560. vitality has
       an add column of 100 and a multiply column of 55. by this metric,
       a weapon with +1,000 vitality is worth as much as a weapon with
       +1 to all skills. this function is already unbalanced, but that type
       of stuff is too obvious to let slide. */
    /* 83: '+1 to Druid Skills', 85 = '+5% Experience Gained',
       76: '+5% Maximum Health', 77 = '+5% Maximum Mana' */
    if(prop->id == 83 || prop->id == 85 || prop->id == 76 || prop->id == 77) {
      val /= 5; }
    /* 87: 'Reduced Vendor Prices' */
    else if(prop->id == 87) {
      val /= 10; }
    /* some of these properties are also insanely low. 141 is deadly strike
       and is highly valued by melee, but has an add col of 31 and a mult col
       of 25.*/
    else if(prop->id == 141) {
      val *= 10; }
    /* 80 = magic find, 93 = ias */
    else if(prop->id == 80 || prop->id == 93) {
      val *= 5; }
    /* 105 = fcr */
    else if(prop->id == 105) {
      val *= 3; }
  }
  /* all prop values are capped at 40. */
  return val > 40 ? 40 : val;
}

/* Checks an item's property lists to determine an estimated item value.
   Again, arbitrary as hell. */
double calc_item_value(struct item_t *item) {
  struct prop_t *prop;
  unsigned int pos;
  double val = 0.0;
  for(prop = item->norm_pl; prop != NULL; prop = prop->next) {
    val += calc_prop_value(prop);
  }
  if(item->qual == set_q) {
    for(pos = 0; pos < 5; pos++) {
      if(item->has_set_pl[pos]) {
	for(prop = item->set_pl[pos]; prop != NULL; prop = prop->next) {
	  val += calc_prop_value(prop);
	}
      }
    }
  }
  return val;
}

/* Determines the 'actual' AC (armor, defense, whatever) equipped on the
   character, when taking proplist modifiers into consideration. */
void calc_ac(struct charinfo_t *charinfo) {
  unsigned int item_no, ac = 0;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    if(is_equipped(charinfo->items[item_no], charinfo->weapon_set) == yes && charinfo->items[item_no]->ac) {
      ac += charinfo->items[item_no]->ac; }
  }
  (charinfo->bonus_stats).def = ac;
}

/* Determines the 'actual' block equipped on the character, when taking
   proplist modifiers into consideration. I guess this doesn't check to see
   whether or not the character actually has a shield equipped... but since
   assassins can block with claws, that's no problem. */
void calc_block(struct charinfo_t *charinfo) {
  unsigned int item_no, block = 0;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    if(is_equipped(charinfo->items[item_no], charinfo->weapon_set) == yes && charinfo->items[item_no]->block) {
      block += charinfo->items[item_no]->block; }
  }
  (charinfo->bonus_stats).block = block;
}

/* Calculates total 'item value' for a character. The item value is computed
   in calc_item_value(), then multiplied by a factor depending on its quality,
   then multiplied by its ilvl. Arbitrary as hell, again. An unused 'rune
   value', as well as the highest carried rune / high rune count, is also
   calculated for list purposes. */
void calc_ival(struct charinfo_t *charinfo) {
  unsigned int item_no, rune_id, ilvl, high_rune = 0, high_rune_count = 0;
  double ival = 0.0, rune_val = 0.0, curr_ival = 0.0;
  char *tag;
  for(item_no = 0; item_no < charinfo->item_count; item_no++) {
    curr_ival = calc_item_value(charinfo->items[item_no]);
    tag = charinfo->items[item_no]->tag;
    ilvl = charinfo->items[item_no]->ilvl;
    /* Some items, like runes, have no ilvl (since they're simple items).
       In this case, either use the required level, or 1. */
    if(!ilvl) {
      ilvl = charinfo->items[item_no]->req_lvl ? charinfo->items[item_no]->req_lvl : 1; }
    /* If item is a rune, check to see if it's the highest level rune that the
       character is carrying (of the ones that have been checked so far). Also
       add to the rune value. */
    if(tag[0] == 'r' && tag[1] >= '0' && tag[1] <= '9' && tag[2] >= '0' && tag[2] <= '9') {
      /* rune tags are always of the form 'rXX', where XX is the rune ID */
      rune_id = tag[2] - '0' + (tag[1] - '0') * 10;
      /* hel runes have no required level, so, give it one. */
      if(rune_id == 15) {
	ilvl = 33; }
      rune_val += ilvl * rune_id;
      if(high_rune < rune_id) {
	high_rune = rune_id;
	high_rune_count = 1; }
      else if(high_rune == rune_id) {
	high_rune_count++; }
    }
    curr_ival *= charinfo->items[item_no]->ilvl;
    /* again, these numbers are pretty arbitrary, you won't find them in any
       game table */
    switch(charinfo->items[item_no]->qual) {
    case unique_q: curr_ival *= 1.5; break;
    case set_q: curr_ival *= 1.3; break;
    case rare_q: curr_ival *= 1.2; break;
    case crafted_q: curr_ival *= 1.3; break;
    default: break;
    }
    ival += curr_ival;
  }
  (charinfo->bonus_stats).ival = ival;
  (charinfo->bonus_stats).rune_val = rune_val;
  (charinfo->bonus_stats).high_rune = (uint8)high_rune;
  (charinfo->bonus_stats).high_rune_count = high_rune_count;
}

/* Calculates the character's 'highest set', checks the percentages of
   set items equipped / total items in the set and assigns the highest
   percentage. This is used later for list purposes. */
void calc_high_set(struct charinfo_t *charinfo) {
  unsigned int pos, high_set;
  double set_val = 0.0, this_val;
  struct set_entry_t *entry;
  for(pos = 0; pos < sizeof_set; pos++) {
    if((entry = lookup_set_by_sid(pos))) {
      this_val = count_bits(charinfo->set_info[pos]) / (double)entry->items;
      if(this_val > set_val) {
	high_set = pos;
	set_val = this_val;
      }
    }
  }
  if(set_val) {
    (charinfo->bonus_stats).high_set = high_set;
    (charinfo->bonus_stats).set_val = set_val; }
}

/* Figures out which ears in the 'ear_list' hashtable belong to which
   characters, tallying the amount of ears each character has floating around
   in the server for list purposes. */
void link_ears_to_owners(struct list_t *list, struct charinfo_t **characters, const unsigned int character_count) {
  struct hashtable_itr *itr;
  unsigned int char_no, val, ret = 1;
  char *key;
  if(list->ear_list) {
    itr = hashtable_iterator(list->ear_list);
    if(itr) {
      for(;ret; ret = hashtable_iterator_advance(itr)) {
	key = (char *)hashtable_iterator_key(itr);
	val = *(unsigned int *)hashtable_iterator_value(itr);
	for(char_no = 0; char_no < character_count; char_no++) {
	  if(characters[char_no] && !strncmp(characters[char_no]->name, key, strlen(key) + 1)) {
	    (characters[char_no]->bonus_stats).ears_dropped = val;
	    break;
	  }
	}
      }
    }
    free(itr);
  }
}
