#include <math.h>
#include <stdlib.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"
#include "tally.h"

/* Combines properties for all items equipped by the character into one
   big proplist for total purposes. */
struct prop_t *generate_total_props(const struct charinfo_t *charinfo) {
  struct prop_t *out = NULL, **prop;
  struct item_t *item;
  struct set_entry_t *set;
  unsigned int item_no, set_no, pos, set_id, set_count, bits_set;
  int bonus = -1; /* more sentinels... */
  for(item_no = 0, prop = &out; item_no < charinfo->item_count; item_no++) {
    item = charinfo->items[item_no];
    if(is_equipped(item, charinfo->weapon_set) == yes) {
      if(item->norm_pl) {
	prop = prop_deep_copy(prop, item->norm_pl); }
      if(item->qual == set_q) {
	set_id = lookup_set_id_by_suid(item->suid);
	bits_set = count_bits(charinfo->set_info[set_id]);
	for(pos = 0; pos < 5; pos++) {
	  if(item->has_set_pl[pos]) {
	    if(bits_set > pos + 1) {
	      prop = prop_deep_copy(prop, item->set_pl[pos]);
	    }
	  }
	}
      }
    }
  }
  /* global set bonuses */
  for(set_no = 0; set_no < sizeof_set; set_no++, bonus = -1) {
    if(charinfo->set_info[set_no]) {
      set = lookup_set_by_sid(set_no);
      set_count = count_bits(charinfo->set_info[set_no]);
      /* determine which global set bonus the player should receive */
      if(set_count == set->items) {
	bonus = 4; }
      else {
	for(pos = 0; pos < 4; pos++) {
	  if(set->set_bonus[pos] && set_count > pos + 1) {
	    bonus = pos; }
	}
      }
      if(bonus != -1) {
	prop = prop_deep_copy(prop, set->set_bonus[bonus]); }
    }
  }
  compile_props(&out);
  return out;
}

/* Uses the total proplist to generate total bonus statistics for the
   character. Also deletes a lot of unnecessary, redundant or inapplicable
   properties so that they don't show up in the bonus list. */
void tally_props(struct prop_t **total, struct charinfo_t *charinfo) {
  struct prop_t **curr, *prop;
  struct bonus_stats_t *bonus = &charinfo->bonus_stats;
  unsigned int tmp;
  int skill_offs;
  zero_out(bonus, sizeof(struct bonus_stats_t));
  for(curr = total; *curr != NULL;) {
    prop = *curr;
    /* Again I'll give an example of the property ID */
    switch(prop->id) {
    case 0:
      /* '+10 to Strength' */
      bonus->str += prop->val;
      delete_prop(curr);
      break;
    case 1:
      /* '+10 to Energy' - this also gives bonus mana */
      bonus->ener += prop->val;
      bonus->mana += lookup_mana_from_ener(prop->val, charinfo->char_class);
      delete_prop(curr);
      break;
    case 2:
      /* '+10 to Dexterity' */
      bonus->dex += prop->val;
      delete_prop(curr);
      break;
    case 3:
      /* '+10 to Vitality' - this also gives bonus life and stam */
      bonus->vit += prop->val;
      bonus->life += lookup_life_from_vit(prop->val, charinfo->char_class);
      bonus->stam += lookup_stam_from_vit(prop->val, charinfo->char_class);
      delete_prop(curr);
      break;
    case 7:
      /* '+50 to Life' */
      bonus->life += prop->val;
      delete_prop(curr);
      break;
    case 9:
      /* '+50 to Mana' */
      bonus->mana += prop->val;
      delete_prop(curr);
      break;
    case 11:
      /* '+50 to Stamina' */
      bonus->stam += prop->val;
      delete_prop(curr);
      break;
    case 19:
      /* '+200 to Attack Rating' */
      bonus->ar += prop->val;
      delete_prop(curr);
      break;
    case 39:
      /* 'Fire Resist +20%' */
      bonus->fire += prop->val;
      delete_prop(curr);
      break;
    case 40:
      /* '+5% to Maximum Fire Resist' */
      bonus->max_fire += prop->val;
      delete_prop(curr);
      break;
    case 41:
      /* 'Lightning Resist +20%' */
      bonus->light += prop->val;
      delete_prop(curr);
      break;
    case 42:
      /* '+5% to Maximum Lightning Resist' */
      bonus->max_light += prop->val;
      delete_prop(curr);
      break;
    case 43:
      /* 'Cold Resist +20%' */
      bonus->cold += prop->val;
      delete_prop(curr);
      break;
    case 44:
      /* '+5% to Maximum Cold Resist' */
      bonus->max_cold += prop->val;
      delete_prop(curr);
      break;
    case 45:
      /* 'Poison Resist +20%' */
      bonus->pois += prop->val;
      delete_prop(curr);
      break;
    case 46:
      /* '+5% to Maximum Poison Resist' */
      bonus->max_pois += prop->val;
      delete_prop(curr);
      break;
    case 60:
      /* '3% Life Stolen Per Hit' */
      bonus->ll += prop->val;
      delete_prop(curr);
      break;
    case 62:
      /* '3% Mana Stolen Per Hit' */
      bonus->ml += prop->val;
      delete_prop(curr);
      break;
    case 76:
      /* 'Increase Maximum Life 10%' */
      bonus->mult_life += prop->val;
      delete_prop(curr);
      break;
    case 77:
      /* 'Increase Maximum Mana 10%' */
      bonus->mult_mana += prop->val;
      delete_prop(curr);
      break;
    case 80:
      /* '50% Better Chance of Getting Magic Items' */
      bonus->mf += prop->val;
      delete_prop(curr);
      break;
    case 83:
      /* '+1 to Barbarian Skill Levels' */
      if(compare_class(prop->param, charinfo->char_class) == yes) {
	bonus->skill0 += prop->val;
	delete_prop(curr); }
      else {
	curr = &(*curr)->next; }
      break;
      /* '20% Increased Attack Speed' */
    case 93:
      bonus->ias += prop->val;
      delete_prop(curr);
      break;
    case 97:
      /* '+20 to Meteor' */
      skill_offs = prop->skill_id - class_skill_offs(charinfo->char_class); 
      if(skill_offs >= 0 && skill_offs < 30) {
	/* if the skill can be learned by the class, the most a
	   non-class-skill prop can give is +3 */
	bonus->skills[skill_offs] += prop->skill_level > 3 ? 3 : prop->skill_level;
	delete_prop(curr); }
      else {
	curr = &(*curr)->next; }
      break;
    case 102:
      /* '+50% Faster Block Speed' */
      bonus->block_speed += prop->val;
      delete_prop(curr);
      break;
    case 105:
      /* '+20% Faster Casting Rate' */
      bonus->fcr += prop->val;
      delete_prop(curr);
      break;
    case 107:
      /* '+2 to Claw Mastery (Assassin Only)' */
      skill_offs = prop->skill_id - class_skill_offs(charinfo->char_class); 
      if(skill_offs >= 0 && skill_offs < 30) {
	bonus->skills[skill_offs] += prop->skill_level;
	delete_prop(curr); }
      else {
	curr = &(*curr)->next; }
      break;
    case 115:
      /* 'Ignore Target's Defense' */
      bonus->has_ignore = yes;
      delete_prop(curr);
      break;
    case 119:
      /* '+5% Bonus to Attack Rating' */
      bonus->mult_ar += prop->val;
      delete_prop(curr);
      break;
    case 127:
      /* '+2 to All Skills' */
      bonus->skill0 += prop->val;
      delete_prop(curr);
      break;
    case 135:
      /* '+50% Chance of Open Wounds' */
      bonus->ow += prop->val;
      delete_prop(curr);
      break;
    case 136:
      /* '+50% Chance of Crushing Blow' */
      bonus->cb += prop->val;
      delete_prop(curr);
      break;
    case 141:
      /* '+50% Chance of Deadly Strike' */
      bonus->ds += prop->val;
      delete_prop(curr);
      break;
      /* '+2 to Offensive Auras (Paladin Only)' */
    case 188:
      if(compare_class((prop->param >> 3 & 0x1ff), charinfo->char_class) == yes) {
	switch(prop->param & 0x7) {
	case 0: bonus->skill1 += prop->val; break;
	case 1: bonus->skill2 += prop->val; break;
	case 2: bonus->skill3 += prop->val; break;
	default: break; }
	delete_prop(curr); }
      else {
	curr = &(*curr)->next; }
      break;
      /* '+50 Life (based on character level)' */
    case 216:
      bonus->life += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 217:
      /* '+50 Mana (based on character level)' */
      bonus->mana += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 220:
      /* '+20 Strength (based on character level)' */
      bonus->str += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 221:
      /* '+20 Dexterity (based on character level)' */
      bonus->dex += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 222:
      /* '+20 Energy (based on character level)' */
      tmp = (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      bonus->ener += tmp;
      bonus->mana += lookup_mana_from_ener(tmp, charinfo->char_class);
      delete_prop(curr);
      break;
    case 223:
      /* '+20 Vitality (based on character level)' */
      tmp = (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      bonus->vit += tmp;
      bonus->life += lookup_life_from_vit(tmp, charinfo->char_class);
      bonus->stam += lookup_stam_from_vit(tmp, charinfo->char_class);
      delete_prop(curr);
      break;
    case 224:
      /* '+1080 to Attack Rating (based on character level)' */
      bonus->ar += (int)floor(((double)prop->val * 0.5 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 225:
      /* '+10% Bonus to Attack Rating (based on character level)' */
      bonus->mult_ar += (int)floor(((double)prop->val * 0.5 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 230:
      /* 'Cold Resist +20% (based on character level)' */
      bonus->cold += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 231:
      /* 'Fire Resist +20% (based on character level)' */
      bonus->fire += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 232:
      /* 'Lightning Resist +20% (based on character level)' */
      bonus->light += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 233:
      /* 'Poison Resist +20% (based on character level)' */
      bonus->pois += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 240:
      /* '20% Better Chance of Getting Magic Items
	 (based on character level)' */
      bonus->mf += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 247:
      /* '+20% Chance of Crushing Blow (based on character level)' */
      bonus->cb += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 248:
      /* '+20% Chance of Open Wounds (based on character level)' */
      bonus->ow += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    case 250:
      /* '+20% Chance of Deadly Strike (based on character level)' */
      bonus->ds += (int)floor(((double)prop->val * 0.125 * (charinfo->stats).level) + 0.5);
      delete_prop(curr);
      break;
    /* All of the following property IDs are in some way insuitable for a
       total list of bonuses that a character gets, and are filtered out. */
    case 16: /* armor percent, not character-global */
    case 17: /* max dmg percent */
    case 18: /* min dmg percent, neither are tallied or displayed */
    case 20: /* +block, already covered with calc_block() */
    case 21: /* min 1h dmg */
    case 22: /* max 1h dmg */
    case 23: /* min 2h dmg */
    case 24: /* max 2h dmg, damage isn't tallied here */
    case 25: /* percent dmg */
    case 31: /* +defense, calc_ac() handles this */
    case 48: /* fire min dmg */
    case 49: /* fire max dmg */
    case 50: /* lightning min dmg */
    case 51: /* lightning max dmg*/
    case 52: /* 'magic' min dmg */
    case 53: /* 'magic' max dmg */
    case 54: /* cold min dmg */
    case 55: /* cold max dmg */
    case 56: /* cold length */
    case 57: /* poison min dmg */
    case 58: /* poison max dmg */
    case 59: /* poison length, all of the above are weapon damage-related */
    case 73: /* max durability */
    case 75: /* % max durability, not character global, and who cares */
    case 91: /* 'Requirements -20%', specific to an item */
    case 92: /* item_levelreq ... ? does this even show up */
    case 94: /* item_levelreqpct ... again, dunno */
    case 125: /* item_throwable */
    case 152: /* 'Indestructible' ... yeah, no you aren't */
    case 159: /* min throw dmg */
    case 160: /* max throw dmg */
    case 214: /* armor per level */
    case 215: /* armor % per level, covered by calc_ac() */
    case 218: /* min dmg per level */
    case 219: /* max dmg per level */
    case 226: /* cold dmg per level */
    case 227: /* fire dmg per level */
    case 228: /* lightning dmg per level */
    case 229: /* poison dmg per level */
    case 252: /* 'Repairs 1 durability every 20 seconds', item-specific */
    case 253: /* 'Replenishes quantity' */
    case 254: /* 'Increased Stack Size' */
      delete_prop(curr);
      break;
    default:
      curr = &(*curr)->next;
      break;
    }
  }
}
