#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "hashtable.h"
#include "tbl.h"

/* Computes a cstring's hash value. Called by hashtable. */
unsigned int hash_cstr(void *vstr) {
  unsigned long hash = 5381; /* magic number */
  char *str = (char *)vstr;
  int c;
  while ((c = *str++)) {
    hash = ((hash << 5) + hash) + c; /* hash * 33 (magic) + c. */
  }
  return hash;
}

/* Compares two cstrings for hashing purposes. 1 means equal, 0 means not. */
int hash_cmp_cstr(void *l, void *r) {
  return !strcmp((const char *)l, (const char *)r);
}

/* Reads a Diablo II ASCII table into a table structure. */
unsigned int read_table(const char *filename, struct table_t *table) {
  char *big_block;
  unsigned int filesize, row, col, pos, chunk;
  FILE *tblfile;
  if(!(tblfile = fopen(filename, "r"))) {
    return 0; }
  if(!(filesize = get_filesize(tblfile))) {
    fclose(tblfile);
    return 0; }
  /* the idea: malloc one big block, read the entire table file raw into this,
     add terminators in the right spots and use char pointers to the entries
     for columns and entries and such. */
  if(!(big_block = malloc(filesize))) {
    fclose(tblfile);
    return 0; }
  if(!(read_bytes(tblfile, (uint8 *)big_block, 0, filesize))) {
    fclose(tblfile);
    return 0; }
  /* check the number of columns. */
  for(pos = 0, table->cols = 0; pos < filesize && big_block[pos] != '\n'; pos++) {
    if(big_block[pos] == '\t') {
      table->cols++; }
  }
  table->cols++;
  /* check the number of rows. */
  for(pos = 0, table->rows = 0; pos < filesize; pos++) {
    if(big_block[pos] == '\n') {
      table->rows++; }
  }
  table->rows--;
  if(!(table->header = malloc(table->cols * sizeof(char *)))) {
    fclose(tblfile);
    free(big_block);
    return 0;
  }
  /* read the header row. */
  for(chunk = pos = col = 0; pos < filesize && col < table->cols; pos++) {
    if(big_block[pos] == '\t' || big_block[pos] == '\n' || big_block[pos] == '\r') {
      table->header[col] = big_block + chunk;
      chunk = pos + 1;
      big_block[pos] = '\0';
      col++;
    }
  }
  if(!(table->entries = malloc(table->rows * sizeof(char **)))) {
    fclose(tblfile);
    free(big_block);
    return 0;
  }
  /* adjust in case of \r\n line terminators. */
  if(big_block[pos] == '\n') {
    pos++; }
  /* read in data, one row at a time. */
  for(row = 0; pos < filesize && row < table->rows; row++) {
    if(!(table->entries[row] = malloc(table->cols * sizeof(char *)))) {
      fclose(tblfile);
      free(big_block);
      return 0;
    }
    for(chunk = pos, col = 0; pos < filesize && col < table->cols; pos++) {
      if(big_block[pos] == '\t' || big_block[pos] == '\n' || big_block[pos] == '\r') {
	table->entries[row][col] = big_block + chunk;
	chunk = pos + 1;
	big_block[pos] = '\0';
	col++;
      }
    }
    /* again, adjust in case of \r\n */
    if(pos < filesize && big_block[pos] == '\n') {
      pos++; }
  }
  return table->rows * table->cols;
}

/* Frees a table structure. */
void free_table(struct table_t *table) {
  unsigned int row;
  if(table) {
    for(row = 0; row < table->rows; row++) {
      free(table->entries[row]);
    }
    if(table->entries) {
      free(table->entries); }
    if(table->header) {
      /* table->header[0] should point to 'big_block' in the function above. */
      free(table->header[0]);
      free(table->header);
    }
  }
}

/* For debugging purposes, print an entire table's rows and two of its
   columns. not actually used, so not commented. */
void print_col(const struct table_t *table, const int col, const int col2) {
  unsigned int row;
  if(col > (int)table->cols || col2 > (int)table->cols || col < 0 || col2 < 0) {
    return; }
  printf("{");
  for(row = 0; row < table->rows; row++) {
    printf(row == table->rows - 1 ? "\"%s\":\"%s\"};\n" : "\"%s\":\"%s\",\n", table->entries[row][col], table->entries[row][col2]); }
}

/* The rest of the functions are basically the same, with a few minor changes.
   Each of these functions has a static hashtable or 2-dim array, and if it
   hasn't been initialized, it reads it in from its corresponding file. If
   it couldn't be read, a sentinel value is assigned to the pointer to this
   static hash/array. Otherwise, it adjusts the input if necessary and returns
   the lookup, provided it exists. */

/* Looks up an item entry by its 3-character tag found in the JM item format.
   files used: 'armor.txt', 'weapons.txt', 'misc.txt'
   sample entry: 'r30' -> 'Ber Rune' */
struct entry_t *lookup_item_by_tag(char *tag) {
  static struct hashtable *lookup = NULL;
  if(!lookup) {
    if(!(lookup = make_item_hash())) {
      write_log("failed to obtain item table\n", 2);
      lookup = (struct hashtable *)1;
      return NULL;
    }
  }
  if(lookup == (struct hashtable *)1) {
    return NULL; }
  else {
    return (struct entry_t *)hashtable_search(lookup, tag); }
}

/* Looks up a property entry by its 9-bit proplist ID.
   files used: 'ItemStatCost.txt'
   sample entry: 80 -> 'item_magicbonus' (MF) */
struct property_entry_t *lookup_prop_by_id(unsigned int id) {
  static struct property_entry_t **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_prop_arr(&size))) {
      write_log("failed to obtain property table\n", 2);
      lookup = (struct property_entry_t **)1;
      return NULL;
    }
  }
  if(lookup == (struct property_entry_t **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a skill entry by its skill ID, which appears in some properties.
   files used: 'skilldesc.txt'
   sample entry: 106 -> 'Zeal' */
struct skill_entry_t *lookup_skill_by_id(unsigned int id) {
  static struct skill_entry_t **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_skill_arr(&size))) {
      write_log("failed to obtain skill table\n", 2);
      lookup = (struct skill_entry_t **)1;
      return NULL;
    }
  }
  if(lookup == (struct skill_entry_t **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a unique item entry by its suid, found in the JM item format.
   files used: 'UniqueItems.txt'
   sample entry: 122 -> 'The Stone of Jordan' */
struct unique_entry_t *lookup_unique_by_id(unsigned int id) {
  static struct unique_entry_t **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_unique_arr(&size))) {
      write_log("failed to obtain unique item table\n", 2);
      lookup = (struct unique_entry_t **)1;
      return NULL;
    }
  }
  if(lookup == (struct unique_entry_t **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a set item entry by its suid, found in the JM item format.
   files used: 'SetItems.txt'
   sample entry: 4 -> 'Hsarus' Iron Fist' */
struct set_item_entry_t *lookup_set_item_by_id(unsigned int id) {
  static struct set_item_entry_t **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_set_item_arr(&size))) {
      write_log("failed to obtain set item table\n", 2);
      lookup = (struct set_item_entry_t **)1;
      return NULL;
    }
  }
  if(lookup == (struct set_item_entry_t **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a prefix name by its ID, found in the JM item format.
   files used: 'MagicPrefix.txt'
   sample entry: 1 -> 'Sturdy' */
char *lookup_magic_pre_by_id(unsigned int id) {
  static char **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_magic_pre_arr(&size))) {
      write_log("failed to obtain magic prefix table\n", 2);
      lookup = (char **)1;
      return NULL;
    }
  }
  if(lookup == (char **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a suffix name by its ID, found in the JM item format.
   files used: 'MagicSuffix.txt'
   sample entry: 5 -> 'of Warding' */
char *lookup_magic_suf_by_id(unsigned int id) {
  static char **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_magic_suf_arr(&size))) {
      write_log("failed to obtain magic suffix table\n", 2);
      lookup = (char **)1;
      return NULL;
    }
  }
  if(lookup == (char **)1) {
    return NULL; }
  else {
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a rare name1 by its ID, found in the JM item format.
   files used: 'RarePrefix.txt'
   sample entry: 0xa3 -> 'Blood' */
char *lookup_rare_pre_by_id(unsigned int id) {
  static char **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_rare_pre_arr(&size))) {
      write_log("failed to obtain rare prefix table\n", 2);
      lookup = (char **)1;
      return NULL;
    }
  }
  if(lookup == (char **)1) {
    return NULL; }
  else {
    id -= 0x9c; /* table offset */
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a rare name2 by its ID, found in the JM item format.
   files used: 'RareSuffix.txt'
   sample entry: 4 -> 'Fang' */
char *lookup_rare_suf_by_id(unsigned int id) {
  static char **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_rare_suf_arr(&size))) {
      write_log("failed to obtain rare suffix table\n", 2);
      lookup = (char **)1;
      return NULL;
    }
  }
  if(lookup == (char **)1) {
    return NULL; }
  else {
    id -= 1; /* your guess is as good as mine, but hey, it works */
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}

/* Looks up a runeword by its runeword ID, found in the JM item format.
   files used: 'Runes.txt'
   sample entry: 80 -> 'Holy Thunder' */
char *lookup_rune_by_id(unsigned int id) {
  static char **lookup = NULL;
  static unsigned int size = 0;
  if(!lookup) {
    if(!(lookup = make_rune_arr(&size))) {
      write_log("failed to obtain rune table\n", 2);
      lookup = (char **)1;
      return NULL;
    }
  }
  if(lookup == (char **)1) {
    return NULL; }
  else {
    /* runeword IDs are ... weird. */
    if(id == 2718) {
      return "Runeword22"; }
    else if(id == 0xc3) {
      return "Runeword170"; }
    /* this seems to work. why? no clue, the table doesn't hint at it */
    id -= id >= 106 ? 25 : 26;
    if(id > size) {
      return NULL; }
    return lookup[id]; }
}
