#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "tbl.h"
#include "hashtable.h"

/* Before, filenames were hardcoded, so this made sense. Now it's kind of
   a relic, but it's still used to determine what file is being read. */
const enum item_type_t type_of_item_in_file[] = {armor, weapon, other};

/* Hardcoded column list for armor.txt, weapons.txt, misc.txt - some will
   not appear in all tables. */
#define sizeof_item_column_list 22
const char *item_column_list[] = {
  "code", "name", "type", "invwidth", "invheight", "cost", "rarity", "levelreq", "reqstr", "reqdex", "mindam", "2handmindam", "minmisdam", "maxdam", "2handmaxdam", "maxmisdam", "gemsockets", "nodurability", "stackable", "quest", "maxac", "block" };

/* Static lookup function that determines which column numbers match which
   column names above. Returns the entry for ('row', 'item_column_list[col]'),
   or NULL if the column hasn't been found. */
char *lookup_item_cols(const struct table_t *table, const unsigned int row, const unsigned int col) {
  static unsigned int col_nums[sizeof_item_column_list], set = 0;
  unsigned int col_find, pos;
  if(table) {
    if(!set) {
      for(pos = 0; pos < table->cols; pos++) {
	for(col_find = 0; col_find < sizeof_item_column_list; col_find++) {
	  if(!strncmp(table->header[pos], item_column_list[col_find], strlen(item_column_list[col_find]) + 1)) {
	    col_nums[col_find] = pos; }
	}
      }
      set = 1;
    }
    if(col_nums[col] == -1) {
      return NULL; }
    return table->entries[row][col_nums[col]];
  }
  else {
    for(col_find = 0; col_find < sizeof_item_column_list; col_find++) {
      col_nums[col_find] = -1;
    }
    set = 0;
    return NULL;
  }
}

/* Determines what item entry is what type (weapon, armor, tome, body, charm,
   gold, other). Not all of these types are used. */
void set_entry_item_type(struct entry_t *entry, const enum item_type_t type) {
  switch(type) {
  case weapon: entry->type = weapon; return;
  case armor: entry->type = armor; return;
  case other:
    if(!strncmp(entry->type_tag, "book", 5) || !strncmp(entry->type_tag, "scro", 5)) {
      entry->type = tome; }
    else if(!strncmp(entry->type_tag, "body", 5)) {
      entry->type = body; }
    else if(!strncmp(entry->type_tag, "scha", 5) || !strncmp(entry->type_tag, "mcha", 5) || !strncmp(entry->type_tag, "lcha", 5)) {
      entry->type = charm; }
    else if(!strncmp(entry->type_tag, "gold", 5)) {
      entry->type = gold; }
    else {
      entry->type = other; }
    return;
  default: return;
  }
}

/* Searches a table's row and mallocs an item entry to match the row. */
struct entry_t *get_item_entry(const struct table_t *table, const unsigned int row) {
  struct entry_t *entry;
  char *str, *newstr;
  int mindmg = 0, maxdmg = 0, tmp = 0;
  if(!(entry = malloc(sizeof(struct entry_t)))) {
    return NULL; }
  zero_out(entry, sizeof(struct entry_t));
  /* for a list of these column numbers, just look at the top of this file */
  if(!(str = lookup_item_cols(table, row, 1)) || str[0] == '\0') {
    return NULL; }
  if(!(newstr = malloc(strlen(str) + 1))) {
    return NULL; }
  strncpy(newstr, str, strlen(str) + 1);
  entry->name = newstr;
  if((str = lookup_item_cols(table, row, 2)) && str[0] != '\0') {
    if(!(newstr = malloc(strlen(str) + 1))) {
      return NULL; }
    strncpy(newstr, str, strlen(str) + 1);
    entry->type_tag = newstr;
  }
  if((str = lookup_item_cols(table, row, 3)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->width); }
  if((str = lookup_item_cols(table, row, 4)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->height); }
  if((str = lookup_item_cols(table, row, 5)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->cost); }
  if((str = lookup_item_cols(table, row, 6)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->rarity); }
  if((str = lookup_item_cols(table, row, 7)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->req_level); }
  if((str = lookup_item_cols(table, row, 8)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->req_str); }
  if((str = lookup_item_cols(table, row, 9)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->req_dex); }
  if((str = lookup_item_cols(table, row, 10)) && str[0] != '\0') {
    sscanf(str, "%d", &mindmg); }
  if((str = lookup_item_cols(table, row, 11)) && str[0] != '\0') {
    sscanf(str, "%d", &tmp);
    mindmg = tmp > mindmg ? tmp : mindmg; }
  if((str = lookup_item_cols(table, row, 12)) && str[0] != '\0') {
    sscanf(str, "%d", &tmp);
    mindmg = tmp > mindmg ? tmp : mindmg; }
  entry->min_dmg = mindmg;
  if((str = lookup_item_cols(table, row, 13)) && str[0] != '\0') {
    sscanf(str, "%d", &maxdmg); }
  if((str = lookup_item_cols(table, row, 14)) && str[0] != '\0') {
    sscanf(str, "%d", &tmp);
    maxdmg = tmp > maxdmg ? tmp : maxdmg; }
  if((str = lookup_item_cols(table, row, 15)) && str[0] != '\0') {
    sscanf(str, "%d", &tmp);
    maxdmg = tmp > maxdmg ? tmp : maxdmg; }
  entry->max_dmg = maxdmg;
  if((str = lookup_item_cols(table, row, 16)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->max_sockets); }
  if((str = lookup_item_cols(table, row, 17)) && str[0] != '\0') {
    tmp = 0;
    sscanf(str, "%d", &tmp);
    entry->no_dura = tmp ? 1 : 0; }
  if((str = lookup_item_cols(table, row, 18)) && str[0] != '\0') {
    tmp = 0;
    sscanf(str, "%d", &tmp);
    entry->has_quantity = tmp ? 1 : 0; }
  if((str = lookup_item_cols(table, row, 19)) && str[0] != '\0') {
    tmp = 0;
    sscanf(str, "%d", &tmp);
    entry->quest = tmp ? 1 : 0; }
  if((str = lookup_item_cols(table, row, 20)) && str[0] != '\0') {
    tmp = 0;
    sscanf(str, "%d", &tmp);
    entry->has_def = tmp ? 1 : 0; }
  if((str = lookup_item_cols(table, row, 21)) && str[0] != '\0') {
    sscanf(str, "%d", &entry->block); }
  return entry;
}

/* Reads each item table file, makes lots of item entries and stores
   each into a hashtable, which is returned. */
struct hashtable *make_item_hash() {
  unsigned int file_no, row, key_len, keys = 0;
  char *key, *tag, filename[1024];
  struct entry_t *val;
  struct hashtable *out;
  struct table_t table;
  zero_out(&table, sizeof(struct table_t));
  out = create_hashtable(1024, &hash_cstr, &hash_cmp_cstr);
  for(file_no = 0; file_no < 3; file_no++) {
    lookup_item_cols(NULL, 0, 0);
    config(armor_file + file_no, filename);
    if(!(read_table(filename, &table))) {
      write_log("could not open table file %s\n", 2, filename);
      continue; }
    if(lookup_item_cols(&table, 0, 0) == NULL || lookup_item_cols(&table, 0, 1) == NULL) {
      write_log("table file %s is invalid?\n", 2, filename);
      free_table(&table);
      continue; }
    for(row = 0; row < table.rows; row++) {
      if((tag = lookup_item_cols(&table, row, 0))[0] != '\0') {
	if(!hashtable_search(out, tag)) {
	  key_len = (unsigned int)strlen(tag) + 1;
	  if(!(key = malloc(key_len))) {
	    return NULL; }
	  strncpy(key, tag, key_len);
	  if(!(val = get_item_entry(&table, row))) {
	    return NULL; }
	  set_entry_item_type(val, type_of_item_in_file[file_no]);
	  if(!hashtable_insert(out, key, val)) {
	    return NULL; }
	  keys++;
	}
	else {
	  write_log("conflicting entries: (%s, %s) -> (%s, %s)\n", 5, tag, lookup_item_cols(&table, row, 1), tag, hashtable_search(out, tag)); }
      }
    }
    free_table(&table);
  }
  if(!keys) {
    hashtable_destroy(out, 1);
    return NULL; }
  return out;
}
