#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "tbl.h"
#include "hashtable.h"

/* UniqueItems.txt table column list. */
#define sizeof_unique_column_list 6
const char *unique_column_list[] = {
  "index", "code", "rarity", "lvl req", "cost mult", "cost add"};

/* SetItems.txt table column list. */
#define sizeof_set_item_column_list 7
const char *set_item_column_list[] = {
  "index", "item", "set", "rarity", "lvl req", "cost mult", "cost add"};

/* Static lookup function that determines which column numbers match which
   column names above. Returns the entry for ('row',
   'unique_column_list[col]'), or NULL if the column hasn't been found. */
char *lookup_unique_cols(const struct table_t *table, const unsigned int row, const unsigned int col) {
  static unsigned int col_nums[sizeof_unique_column_list], set = 0;
  unsigned int col_find, pos;
  if(table) {
    if(!set) {
      for(pos = 0; pos < table->cols; pos++) {
	for(col_find = 0; col_find < sizeof_unique_column_list; col_find++) {
	  if(!strncmp(table->header[pos], unique_column_list[col_find], strlen(unique_column_list[col_find]) + 1)) {
	    col_nums[col_find] = pos; }
	}
      }
      set = 1;
    }
    if(col_nums[col] == -1) {
      return NULL; }
    return table->entries[row][col_nums[col]];
  }
  else {
    for(col_find = 0; col_find < sizeof_unique_column_list; col_find++) {
      col_nums[col_find] = -1;
    }
    set = 0;
    return NULL;
  }
}

/* Static lookup function that determines which column numbers match which
   column names above. Returns the entry for ('row',
   'set_item_column_list[col]'), or NULL if the column hasn't been found. */
char *lookup_set_item_cols(const struct table_t *table, const unsigned int row, const unsigned int col) {
  static unsigned int col_nums[sizeof_set_item_column_list], set = 0;
  unsigned int col_find, pos;
  if(table) {
    if(!set) {
      for(pos = 0; pos < table->cols; pos++) {
	for(col_find = 0; col_find < sizeof_set_item_column_list; col_find++) {
	  if(!strncmp(table->header[pos], set_item_column_list[col_find], strlen(set_item_column_list[col_find]) + 1)) {
	    col_nums[col_find] = pos; }
	}
      }
      set = 1;
    }
    if(col_nums[col] == -1) {
      return NULL; }
    return table->entries[row][col_nums[col]];
  }
  else {
    for(col_find = 0; col_find < sizeof_set_item_column_list; col_find++) {
      col_nums[col_find] = -1;
    }
    set = 0;
    return NULL;
  }
}

/* Searches a table's row and mallocs a unique item entry to match the row. */
struct unique_entry_t *get_unique_entry(const struct table_t *table, const unsigned int row) {
  struct unique_entry_t *entry;
  char *str, *newstr;
  if(!(entry = malloc(sizeof(struct unique_entry_t)))) {
    return NULL; }
  zero_out(entry, sizeof(struct unique_entry_t));
  if(!(str = lookup_unique_cols(table, row, 0)) || str[0] == '\0') {
    free(entry);
    return NULL; }
  if(!(newstr = malloc(strlen(str) + 1))) {
    return NULL; }
  strncpy(newstr, str, strlen(str) + 1);
  entry->name = newstr;
  if(!(str = lookup_unique_cols(table, row, 1)) || str[0] == '\0') {
    free(newstr);
    free(entry);
    return NULL; }
  if(!(newstr = malloc(strlen(str) + 1))) {
    return NULL; }
  strncpy(newstr, str, strlen(str) + 1);
  entry->type_tag = newstr;
  if((str = lookup_unique_cols(table, row, 2)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->rarity); }
  if((str = lookup_unique_cols(table, row, 3)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->req_level); }
  if((str = lookup_unique_cols(table, row, 4)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->cost_mult); }
  if((str = lookup_unique_cols(table, row, 5)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->cost_add); }
  return entry;
}

/* Searches a table's row and mallocs a set item entry to match the row. */
struct set_item_entry_t *get_set_item_entry(const struct table_t *table, const unsigned int row) {
  struct set_item_entry_t *entry;
  char *str, *newstr;
  if(!(entry = malloc(sizeof(struct set_item_entry_t)))) {
    return NULL; }
  zero_out(entry, sizeof(struct set_item_entry_t));
  if(!(str = lookup_set_item_cols(table, row, 0)) || str[0] == '\0') {
    free(entry);
    return NULL; }
  if(!(newstr = malloc(strlen(str) + 1))) {
    return NULL; }
  strncpy(newstr, str, strlen(str) + 1);
  entry->name = newstr;
  if(!(str = lookup_set_item_cols(table, row, 1)) || str[0] == '\0') {
    free(newstr);
    free(entry);
    return NULL; }
  if(!(newstr = malloc(strlen(str) + 1))) {
    return NULL; }
  strncpy(newstr, str, strlen(str) + 1);
  entry->type_tag = newstr;
  if((str = lookup_set_item_cols(table, row, 2)) && str[0] != '\0') {
    if(!(newstr = malloc(strlen(str) + 1))) {
      return NULL; }
    strncpy(newstr, str, strlen(str) + 1);
    entry->set = newstr;
  }
  if((str = lookup_set_item_cols(table, row, 3)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->rarity); }
  if((str = lookup_set_item_cols(table, row, 4)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->req_level); }
  if((str = lookup_set_item_cols(table, row, 5)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->cost_mult); }
  if((str = lookup_set_item_cols(table, row, 6)) && str[0] != '\0') {
    sscanf(str, "%u", &entry->cost_add); }
  return entry;
}

/* Reads UniqueItems.txt, makes lots of item entries and stores
   each into an array, which is returned. */
struct unique_entry_t **make_unique_arr(unsigned int *out_size) {
  unsigned int row, keys = 0, size = 1024;
  struct unique_entry_t *val, **out;
  struct table_t table;
  char filename[1024];
  zero_out(&table, sizeof(struct table_t));
  if(!(out = malloc(size * sizeof(struct unique_entry_t *)))) {
    return NULL; }
  zero_out(out, size * sizeof(struct unique_entry_t *));
  lookup_unique_cols(NULL, 0, 0);
  config(unique_file, filename);
  if(!(read_table(filename, &table))) {
    write_log("could not open table file %s\n", 2, filename);
    return NULL; }
  if(lookup_unique_cols(&table, 0, 0) == NULL || lookup_unique_cols(&table, 0, 1) == NULL) {
    write_log("table file %s is invalid?\n", 2, filename);
    free_table(&table);
    return NULL; }
  for(row = 0; row < table.rows; row++) {
    if((val = get_unique_entry(&table, row))) {
      if(keys > size) {
	size *= 2;
	if(!(out = realloc(out, size * sizeof(struct unique_entry_t *)))) {
	  free(out);
	  free_table(&table);
	  return NULL;
	}
	zero_out(out + keys, size / 2 * sizeof(struct unique_entry_t *));
      }
      if(out[keys] != NULL) {
	free(out[keys]); }
      out[keys] = val;
      keys++;
    }
    else {
      if(strncmp(lookup_unique_cols(&table, row, 0), "Expansion", 10)) {
	keys++; }
    }
  }
  free_table(&table);
  if(!keys) {
    free(out);
    return NULL; }
  *out_size = keys;
  return out;
}

/* Reads SetItems.txt, makes lots of item entries and stores
   each into an array, which is returned. */
struct set_item_entry_t **make_set_item_arr(unsigned int *out_size) {
  unsigned int row, keys = 0, size = 1024;
  struct set_item_entry_t *val, **out;
  struct table_t table;
  char filename[1024];
  zero_out(&table, sizeof(struct table_t));
  if(!(out = malloc(size * sizeof(struct set_item_entry_t *)))) {
    return NULL; }
  zero_out(out, size * sizeof(struct set_item_entry_t *));
  lookup_set_item_cols(NULL, 0, 0);
  config(set_item_file, filename);
  if(!(read_table(filename, &table))) {
    write_log("could not open table file %s\n", 2, filename);
    return NULL; }
  if(lookup_set_item_cols(&table, 0, 0) == NULL || lookup_set_item_cols(&table, 0, 1) == NULL) {
    write_log("table file %s is invalid?\n", 2, filename);
    free_table(&table);
    return NULL; }
  for(row = 0; row < table.rows; row++) {
    if((val = get_set_item_entry(&table, row))) {
      if(keys > size) {
	size *= 2;
	if(!(out = realloc(out, size * sizeof(struct set_item_entry_t *)))) {
	  free(out);
	  free_table(&table);
	  return NULL;
	}
	zero_out(out + keys, size / 2 * sizeof(struct set_item_entry_t *));
      }
      if(out[keys] != NULL) {
	free(out[keys]); }
      out[keys] = val;
      keys++;
    }
  }
  free_table(&table);
  if(!keys) {
    free(out);
    return NULL; }
  *out_size = keys;
  return out;
}

/* Returns a runeword cstring array, containing all the runeword names
   contained in the file specified by config. Sets out_size to its bound. */
char **make_rune_arr(unsigned int *out_size) {
  unsigned int row, keys = 0, key, size = 512, name_col = -1, col, name_len;
  char *val, **out, filename[1024];
  struct table_t table;
  zero_out(&table, sizeof(struct table_t));
  if(!(out = malloc(size * sizeof(char *)))) {
    return NULL; }
  zero_out(out, size * sizeof(char *));
  config(rune_file, filename);
  if(!(read_table(filename, &table))) {
    write_log("could not open table file %s\n", 2, filename);
    return NULL; }
  for(col = 0; col < table.cols; col++) {
    if(!strncmp(table.header[col], "Name", 5)) {
      name_col = col; }
  }
  if(name_col == -1) {
    write_log("table file %s is invalid?\n", 2, filename);
    free_table(&table);
    return NULL; }
  for(row = 0; row < table.rows; row++) {
    /* Check to make sure the row is valid (has a name column and a scannable
       "RunewordXXX" column), then malloc the value cstring
       (which will contain the runeword name). */
    if(strlen(table.entries[row][name_col]) > 8 && sscanf(table.entries[row][name_col] + 8, "%d", &key)) {
      name_len = (unsigned int)strlen(table.entries[row][name_col]) + 1;
      if(!(val = malloc(name_len))) {
	free(out);
	free_table(&table);
	return NULL; }
      strncpy(val, table.entries[row][name_col], name_len);
      /* If the table has more data than anticipated, resize the array. */
      if(key > size) {
	/* If the table has way data than anticipated, return null. */
	if(key > 1024) {
	  free(out);
	  free_table(&table);
	  return NULL;
	}
	size *= 2;
	if(!(out = realloc(out, size * sizeof(struct skill_entry_t *)))) {
	  free(out);
	  free_table(&table);
	  return NULL;
	}
      }
      /* Free duplicates. */
      if(out[key] != NULL) {
	free(out[key]); }
      out[key] = val;
      keys++;
    }
  }
  free_table(&table);
  if(!keys) {
    free(out);
    return NULL; }
  *out_size = keys;
  return out;
}

/* Simple debug function to print runeword names, not commented. */
void rw_test() {
  int pos = 0;
  char *name, *str;
  for(pos = 0; pos < 200; pos++) {
    if((name = lookup_rune_by_id(pos))) {
      str = lookup_string_by_name(name);
      printf("%d: %s\n", pos, str ? str : name);
    }
  }
}

