#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "d2stats.h"
#include "items.h"
#include "tbl.h"
#include "print.h"
#include "tally.h"
#include "xml.h"

unsigned int indent_level = 0;

/* Basically writes <'tag'> to file, along with indent. */
void xml_open_tag(FILE *file, const char *tag) {
  unsigned int pos;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  fprintf(file, "<%s>", tag);
}

/* Basically writes <'tag' "'attr'"="'val'"> to file, along with indent. */
void xml_open_attr(FILE *file, const char *tag, const char *attr, const char *val) {
  unsigned int pos;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  fprintf(file, "<%s %s=\"%s\">", tag, attr, val);
}

/* Basically writes </'tag'>. */
void xml_close_tag(FILE *file, const char *tag) {
  fprintf(file, "</%s>\n", tag);
}

/* Writes <'tag'> with indent, and increments indent level.*/
void xml_open_tree(FILE *file, const char *tag) {
  unsigned int pos;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  indent_level++;
  fprintf(file, "<%s>\n", tag);
}

/* Writes <'tag' "'attr'"="'val'"> with indent, and increments indent level. */
void xml_open_tree_attr(FILE *file, const char *tag, const char *attr, const char *val) {
  unsigned int pos;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  indent_level++;
  fprintf(file, "<%s %s=\"%s\">\n", tag, attr, val);
}

/* Writes <'tag' "'attr'"="'val'" "'attr2'"="'val2'"> with indent,
   and increments indent level. */
void xml_open_tree_2attr(FILE *file, const char *tag, const char *attr, const char *val, const char *attr2, const char *val2) {
  unsigned int pos;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  indent_level++;
  fprintf(file, "<%s %s=\"%s\" %s=\"%s\">\n", tag, attr, val, attr2, val2);
}

/* Writes </'tag'> with indent and decrements indent level. */
void xml_close_tree(FILE *file, const char *tag) {
  unsigned int pos;
  indent_level--;
  for(pos = 0; pos < indent_level; pos++) {
    fprintf(file, "  "); }
  fprintf(file, "</%s>\n", tag);
}

/* Writes an XML header for character files. */
void xml_header(FILE *file, const char *tag) {
  char buf[1024];
  config(char_xslt, buf);
  fprintf(file, "<?xml version=\"1.0\"?>\n");
  if(buf[0] != '\0') {
    fprintf(file, "<?xml-stylesheet type=\"text/xsl\" href=\"%s\"?>\n", buf); }
  fprintf(file, "<%s xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n", tag);
  fprintf(file, " xsi:schemaLocation=\"http://vidya.dyndns.org/xml/d2stats http://vidya.dyndns.org/xml/d2stats.xsd\"\n");
  fprintf(file, " xmlns=\"http://vidya.dyndns.org/xml/d2stats\"\n");
  fprintf(file, " xmlns:i=\"http://vidya.dyndns.org/xml/items\"\n");
  fprintf(file, " xmlns:c=\"http://vidya.dyndns.org/xml/charinfo\"\n");
  fprintf(file, " xmlns:p=\"http://vidya.dyndns.org/xml/properties\">\n");
  indent_level = 1;
}

/* Writes an XML header for list files. */
void xml_list_header(FILE *file, const char *tag) { 
  char buf[1024];
  config(list_xslt, buf);
 fprintf(file, "<?xml version=\"1.0\"?>\n");
  if(buf[0] != '\0') {
    fprintf(file, "<?xml-stylesheet type=\"text/xsl\" href=\"%s\"?>\n", buf); }
  fprintf(file, "<%s xmlns=\"http://vidya.dyndns.org/xml/d2list\"\n", tag);
  fprintf(file, " xmlns:l=\"http://vidya.dyndns.org/xml/list\">\n");
  indent_level = 1;
}

/* Writes an XML footer for character files. */
void xml_footer(FILE *file, const char *tag) {
  fprintf(file, "</%s>\n", tag);
}

/* Writes an XML footer for list files. */
void xml_list_footer(FILE *file, const char *tag) {
  fprintf(file, "</%s>\n", tag);
}

/* Writes the 'c:chartitle' XML tag, including character name stuff. */
void xml_name(FILE *file, const struct charinfo_t *charinfo) {
  char *title;
  if((title = char_title_to_cstr(charinfo->char_title, charinfo->char_class, charinfo->is_hardcore)) && title[0] != '\0') {
    xml_open_tag(file, "c:chartitle");
    fprintf(file, "%s", title);
    xml_close_tag(file, "c:chartitle"); }
  xml_open_tag(file, "c:name");
  fprintf(file, "%s", charinfo->name);
  xml_close_tag(file, "c:name");
  if(charinfo->is_ladder == yes) {
    xml_open_tag(file, "c:ladder");
    fprintf(file, "yes");
    xml_close_tag(file, "c:ladder"); }
  if(charinfo->is_hardcore == yes) {
    xml_open_tag(file, "c:hardcore");
    fprintf(file, "yes");
    xml_close_tag(file, "c:hardcore"); }
  xml_open_tag(file, "c:has_died");
  fprintf(file, "%s", charinfo->has_died == yes ? "yes" : "no");
  xml_close_tag(file, "c:has_died");
  xml_open_tag(file, "c:class");
  fprintf(file, "%s", char_class_to_cstr(charinfo->char_class));
  xml_close_tag(file, "c:class");
}

/* Writes the bonus stats XML tags. */
void xml_bonus_stats(FILE *file, const struct bonus_stats_t *bonus, const int res_mod) {
  if(bonus->str) {
    xml_open_tag(file, "c:bonus_str");
    fprintf(file, "%d", bonus->str);
    xml_close_tag(file, "c:bonus_str"); }
  if(bonus->dex) {
    xml_open_tag(file, "c:bonus_dex");
    fprintf(file, "%d", bonus->dex);
    xml_close_tag(file, "c:bonus_dex"); }
  if(bonus->vit) {
    xml_open_tag(file, "c:bonus_vit");
    fprintf(file, "%d", bonus->vit);
    xml_close_tag(file, "c:bonus_vit"); }
  if(bonus->ener) {
    xml_open_tag(file, "c:bonus_ener");
    fprintf(file, "%d", bonus->ener);
    xml_close_tag(file, "c:bonus_ener"); }
  if(bonus->life) {
    xml_open_tag(file, "c:bonus_life");
    fprintf(file, "%d", bonus->life);
    xml_close_tag(file, "c:bonus_life"); }
  if(bonus->mult_life) {
    xml_open_tag(file, "c:bonus_life_mult");
    fprintf(file, "%d", bonus->mult_life);
    xml_close_tag(file, "c:bonus_life_mult"); }
  if(bonus->stam) {
    xml_open_tag(file, "c:bonus_stam");
    fprintf(file, "%d", bonus->stam);
    xml_close_tag(file, "c:bonus_stam"); }
  if(bonus->mana) {
    xml_open_tag(file, "c:bonus_mana");
    fprintf(file, "%d", bonus->mana);
    xml_close_tag(file, "c:bonus_mana"); }
  if(bonus->mult_mana) {
    xml_open_tag(file, "c:bonus_mana_mult");
    fprintf(file, "%d", bonus->mult_mana);
    xml_close_tag(file, "c:bonus_mana_mult"); }
  xml_open_tag(file, "c:bonus_fire");
  fprintf(file, "%d/%d", bonus->fire + res_mod, bonus->max_fire + 75 > 95 ? 95 : bonus->max_fire + 75);
  xml_close_tag(file, "c:bonus_fire");
  xml_open_tag(file, "c:bonus_light");
  fprintf(file, "%d/%d", bonus->light + res_mod, bonus->max_light + 75 > 95 ? 95 : bonus->max_light + 75);
  xml_close_tag(file, "c:bonus_light");
  xml_open_tag(file, "c:bonus_cold");
  fprintf(file, "%d/%d", bonus->cold + res_mod, bonus->max_cold + 75 > 95 ? 95 : bonus->max_cold + 75);
  xml_close_tag(file, "c:bonus_cold");
  xml_open_tag(file, "c:bonus_pois");
  fprintf(file, "%d/%d", bonus->pois + res_mod, bonus->max_pois + 75 > 95 ? 95 : bonus->max_pois + 75);
  xml_close_tag(file, "c:bonus_pois");
  if(bonus->def) {
    xml_open_tag(file, "c:bonus_def");
    fprintf(file, "%d", bonus->def);
    xml_close_tag(file, "c:bonus_def"); }
  if(bonus->ar) {
    xml_open_tag(file, "c:bonus_ar");
    fprintf(file, "%d", bonus->ar);
    xml_close_tag(file, "c:bonus_ar"); }
  if(bonus->mult_ar) {
    xml_open_tag(file, "c:bonus_ar_mult");
    fprintf(file, "%d", bonus->mult_ar);
    xml_close_tag(file, "c:bonus_ar_mult"); }
  if(bonus->has_ignore == yes) {
    xml_open_tag(file, "c:bonus_ignore_ac");
    fprintf(file, "yes");
    xml_close_tag(file, "c:bonus_ignore_ac"); }
  if(bonus->block) {
    xml_open_tag(file, "c:bonus_block");
    fprintf(file, "%d", bonus->block);
    xml_close_tag(file, "c:bonus_block"); }
  if(bonus->block_speed) {
    xml_open_tag(file, "c:bonus_block_speed");
    fprintf(file, "%d", bonus->block_speed);
    xml_close_tag(file, "c:bonus_block_speed"); }
  if(bonus->ias) {
    xml_open_tag(file, "c:bonus_ias");
    fprintf(file, "%d", bonus->ias);
    xml_close_tag(file, "c:bonus_ias"); }
  if(bonus->fcr) {
    xml_open_tag(file, "c:bonus_fcr");
    fprintf(file, "%d", bonus->fcr);
    xml_close_tag(file, "c:bonus_fcr"); }
  if(bonus->cb) {
    xml_open_tag(file, "c:bonus_cb");
    fprintf(file, "%d", bonus->cb);
    xml_close_tag(file, "c:bonus_cb"); }
  if(bonus->ds) {
    xml_open_tag(file, "c:bonus_ds");
    fprintf(file, "%d", bonus->ds);
    xml_close_tag(file, "c:bonus_ds"); }
  if(bonus->ow) {
    xml_open_tag(file, "c:bonus_ow");
    fprintf(file, "%d", bonus->ow);
    xml_close_tag(file, "c:bonus_ow"); }
  if(bonus->mf) {
    xml_open_tag(file, "c:bonus_mf");
    fprintf(file, "%d", bonus->mf);
    xml_close_tag(file, "c:bonus_mf"); }
  if(bonus->ll) {
    xml_open_tag(file, "c:bonus_ll");
    fprintf(file, "%d", bonus->ll);
    xml_close_tag(file, "c:bonus_ll"); }
  if(bonus->ml) {
    xml_open_tag(file, "c:bonus_ml");
    fprintf(file, "%d", bonus->ml);
    xml_close_tag(file, "c:bonus_ml"); }
}

/* Writes the 'c:stats' XML tag. */
void xml_stats(FILE *file, const struct stats_t *stats, const struct bonus_stats_t *bonus_stats, const int res_mod) {
  xml_open_tree(file, "c:stats");
  xml_open_tag(file, "c:level");
  fprintf(file, "%d", stats->level);
  xml_close_tag(file, "c:level");
  xml_open_tag(file, "c:exp");
  fprintf(file, "%d", stats->exp);
  xml_close_tag(file, "c:exp");
  xml_open_tag(file, "c:str");
  fprintf(file, "%d", stats->str);
  xml_close_tag(file, "c:str");
  xml_open_tag(file, "c:dex");
  fprintf(file, "%d", stats->dex);
  xml_close_tag(file, "c:dex");
  xml_open_tag(file, "c:vit");
  fprintf(file, "%d", stats->vit);
  xml_close_tag(file, "c:vit");
  xml_open_tag(file, "c:ener");
  fprintf(file, "%d", stats->ener);
  xml_close_tag(file, "c:ener");
  xml_open_tag(file, "c:life");
  fprintf(file, "%d", (int)(stats->health / 256.0)); 
  xml_close_tag(file, "c:life");
  xml_open_tag(file, "c:stam");
  fprintf(file, "%d", (int)(stats->stam / 256.0));
  xml_close_tag(file, "c:stam");
  xml_open_tag(file, "c:mana");
  fprintf(file, "%d", (int)(stats->mana / 256.0));
  xml_close_tag(file, "c:mana");
  xml_open_tag(file, "c:gold");
  fprintf(file, "%d", stats->gold);
  xml_close_tag(file, "c:gold");
  xml_open_tag(file, "c:stash");
  fprintf(file, "%d", stats->gold_stash);
  xml_close_tag(file, "c:stash");
  if(stats->skills_left) {
    xml_open_tag(file, "c:skills_left");
    fprintf(file, "%d", stats->skills_left);
    xml_close_tag(file, "c:skills_left"); }
  if(stats->stats_left) {
    xml_open_tag(file, "c:stats_left");
    fprintf(file, "%d", stats->stats_left);
    xml_close_tag(file, "c:stats_left"); }
  xml_bonus_stats(file, bonus_stats, res_mod);
  xml_close_tree(file, "c:stats");
}

/* Writes the 'c:skills' XML tag. */
void xml_skills(FILE *file, const struct charinfo_t *charinfo) {
  unsigned int offs, pos;
  char *skill_name, buf[64];
  const struct stats_t *stats = &charinfo->stats;
  const struct bonus_stats_t *bonus_stats = &charinfo->bonus_stats;
  const enum char_class_t char_class = charinfo->char_class;
  offs = class_skill_offs(char_class);
  xml_open_tree(file, "c:skills");
  if(bonus_stats->skill0 || bonus_stats->skill1 || bonus_stats->skill2 || bonus_stats->skill3) {
    xml_open_tag(file, "c:bonus_all_skills");
    fprintf(file, "%d", bonus_stats->skill0);
    xml_close_tag(file, "c:bonus_all_skills");
    xml_open_tag(file, "c:bonus_tab1");
    fprintf(file, "%d", bonus_stats->skill1);
    xml_close_tag(file, "c:bonus_tab1");
    xml_open_tag(file, "c:bonus_tab2");
    fprintf(file, "%d", bonus_stats->skill2);
    xml_close_tag(file, "c:bonus_tab2");
    xml_open_tag(file, "c:bonus_tab3");
    fprintf(file, "%d", bonus_stats->skill3);
    xml_close_tag(file, "c:bonus_tab3"); }
  get_spec(charinfo, buf);
  if(strncmp(buf, "0/0/0", 6)) {
    xml_open_tag(file, "c:spec");
    fprintf(file, "%s", buf);
    xml_close_tag(file, "c:spec"); }
  for(pos = 0; pos < 30; pos++) {
    if(stats->skills[pos] || bonus_stats->skills[pos]) {
      xml_open_tree(file, "c:skill");
      if(stats->skills[pos]) {
	xml_open_tag(file, "c:level");
	fprintf(file, "%d", stats->skills[pos]);
	xml_close_tag(file, "c:level"); }
      if(bonus_stats->skills[pos]) {
	xml_open_tag(file, "c:bonus_level");
	fprintf(file, "%d", bonus_stats->skills[pos]);
	xml_close_tag(file, "c:bonus_level"); }
      xml_open_tag(file, "c:name");
      skill_name = "(unknown skill)";
      if(lookup_skill_by_id(pos + offs)) {
	skill_name = lookup_skill_by_id(pos + offs)->str_name;
	if(lookup_string_by_name(skill_name)) {
	  skill_name = lookup_string_by_name(skill_name); }
      }
      fprintf(file, "%s", skill_name);
      xml_close_tag(file, "c:name");
      xml_close_tree(file, "c:skill");
    }
  }
  xml_close_tree(file, "c:skills");
}

/* Writes the 'c:set_bonus' XML tag. */
void xml_set_bonus(FILE *file, const struct charinfo_t *charinfo) {
  unsigned int set_no, set_count, pos;
  int bonus = -1;
  char *name;
  enum flag_t bonus_printed = no;
  struct set_entry_t *set;
  for(set_no = 0; set_no < sizeof_set; set_no++, bonus = -1) {
    if(charinfo->set_info[set_no]) {
      set = lookup_set_by_sid(set_no);
      set_count = count_bits(charinfo->set_info[set_no]);
      if(set_count == set->items) {
	bonus = 4; }
      else {
	for(pos = 0; pos < 4; pos++) {
	  if(set->set_bonus[pos] && set_count > pos + 1) {
	    bonus = pos; }
	}
      }
      if(bonus != -1) {
	if(bonus_printed == no) {
	  bonus_printed = yes;
	  xml_open_tree(file, "c:set_bonus"); }
	name = set->name;
	if(lookup_string_by_name(name)) {
	  name = lookup_string_by_name(name); }
	xml_open_tree(file, "c:set");
	xml_open_tag(file,"c:setname");
	fprintf(file, "%s", name);
	xml_close_tag(file, "c:setname");
	xml_open_tag(file, "c:setcount");
	fprintf(file, "%d / %d", set_count, set->items);
	xml_close_tag(file, "c:setcount");
	xml_open_tree(file, "p:proplist");
	xml_prop_list(file, set->set_bonus[bonus], (charinfo->stats).level);
	xml_close_tree(file, "p:proplist");
	xml_close_tree(file, "c:set");
      }
    }
  }
  if(bonus_printed == yes) {
    xml_close_tree(file, "c:set_bonus"); }
}

/* Writes the 'c:totals' XML tag. */
void xml_totals(FILE *file, struct charinfo_t *charinfo) {
  if(charinfo->total) {
    xml_open_tree(file, "c:totals");
    xml_open_tree(file, "p:proplist");
    xml_prop_list(file, charinfo->total, (charinfo->stats).level);
    xml_close_tree(file, "p:proplist");
    xml_close_tree(file, "c:totals");
  }
}

/* Writes the 'c:charinfo' XML tag. */
void xml_char(FILE *file, struct charinfo_t *charinfo) {
  xml_open_tree(file, "c:charinfo");
  xml_open_tag(file, "c:filename");
  fprintf(file, "%s", charinfo->filename);
  xml_close_tag(file, "c:filename");
  xml_name(file, charinfo);
  xml_stats(file, &charinfo->stats, &charinfo->bonus_stats, get_res_mod(charinfo));
  xml_skills(file, charinfo);
  xml_items(file, charinfo);
  xml_set_bonus(file, charinfo);
  xml_totals(file, charinfo);
  xml_close_tree(file, "c:charinfo");
}

/* Opens a file handle based on config (or stdout if the config isn't set)
   and prints XML character info to it. */
void xml_print(struct charinfo_t *charinfo) {
  char filename[1024] = "";
  FILE *file;
  get_xml_char_file(charinfo, filename);
  if(filename[0] == '\0') {
    file = stdout; }
  else {
    if(!(file = fopen(filename, "w"))) {
      write_log("could not open file \"%s\" for writing\n", 3, filename);
      return; }
  }
  xml_header(file, "d2stats");
  xml_char(file, charinfo);
  xml_footer(file, "d2stats");
  if(file != stdout) {
    fclose(file); }
}
